'''
Created on 20 oct. 2012

@author: coissac
'''

from .build_ctools import build_ctools
from .build_exe import build_exe
from distutils.errors import DistutilsSetupError
from distutils import log
import os

class build_cexe(build_ctools):

    description = "build C/C++ executable distributed with Python extensions"


    def initialize_options(self):
        build_ctools.initialize_options(self)
        self.built_files = None


    def finalize_options(self):
        # This might be confusing: both build-cexe and build-temp default
        # to build-temp as defined by the "build" command.  This is because
        # I think that C libraries are really just temporary build
        # by-products, at least from the point of view of building Python
        # extensions -- but I want to keep my options open.

        build_cexe_dir = self.build_cexe
        build_ctools.finalize_options(self)

        if build_cexe_dir is None:
            self.build_cexe=None
            
        self.set_undefined_options('build',
                                   ('build_scripts',  'build_cexe'))

        self.set_undefined_options('build_files',
                                   ('files',  'built_files'))
                   
        self.executables   = self.distribution.executables
#         self.build_cexe = os.path.join(os.path.dirname(self.build_cexe),'cbinaries') 
#         self.mkpath(self.build_cexe)

        if self.executables:
            self.check_executable_list(self.executables)
                            

        # XXX same as for build_ext -- what about 'self.define' and
        # 'self.undef' ?

    def substitute_sources(self,exe_name,sources):
        """
        Substitutes source file name starting by an @ by the actual
        name of the built file (see --> build_files)
        """
        sources = list(sources)
        for i in range(len(sources)):
            message = "%s :-> %s" % (exe_name,sources[i])
            if sources[i][0]=='@':
                try:
                    filename = self.built_files[sources[i][1:]]
                except KeyError:
                    raise DistutilsSetupError(
                         'The %s filename declared in the source '
                         'files of the program %s have not been '
                         'built by the installation process' % (sources[i],
                                                                exe_name))
                sources[i]=filename
                log.info("%s changed to %s",message,filename)
            else:
                log.info("%s ok",message)

        return sources
 
    def run(self):

        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

        build_exe.run(self)
        

   



#cython: language_level=3, boundscheck=False, wraparound=False

from cpython.list cimport PyList_GET_ITEM,PyList_SET_ITEM,PyList_New

from ._ahocorasick cimport *
from orgasm.apps.progress cimport ProgressBar 

from orgasm.utils.dna import reverseComplement
from ._bitvector cimport BitVector

import time
import math

#
# This dict is a mix off all mitochondrial, chroroplast and the universal genetic code
# It serves for back translating purpose
#
 
cdef dict UniversalGeneticCode     = {b"A" : (b"GCT", b"GCC", b"GCA", b"GCG"),
                                      b"C" : (b"TGT", b"TGC"),
                                      b"D" : (b"GAT", b"GAC"),
                                      b"E" : (b"GAA", b"GAG"),
                                      b"F" : (b"TTT", b"TTC"),
                                      b"G" : (b"GGT", b"GGC", b"GGA", b"GGG"),
                                      b"H" : (b"CAT", b"CAC"),
                                      b"I" : (b"ATT", b"ATC", b"ATA"),
                                      b"K" : (b"AAA", b"AAG"),
                                      b"L" : (b"TTA", b"TTG", b"TAG", b"CTT", b"CTC", b"CTA", b"CTG"),
                                      b"M" : (b"ATG", b"ATA"),
                                      b"N" : (b"AAT", b"AAC", b"AAA"),
                                      b"P" : (b"CCT", b"CCC", b"CCA", b"CCG"),
                                      b"Q" : (b"CAA", b"CAG"),
                                      b"R" : (b"CGT", b"CGC", b"CGA", b"CGG", b"AGA", b"AGG"),
                                      b"S" : (b"TCT", b"TCC", b"TCA", b"TCG", b"AGT", b"AGC", b"AGA", b"AGG"),
                                      b"T" : (b"ACT", b"ACC", b"ACA", b"ACG", b"CTT", b"CTC", b"CTA", b"CTG"),
                                      b"V" : (b"GTT", b"GTC", b"GTA", b"GTG"),
                                      b"W" : (b"TGA", b"TGG"),
                                      b"Y" : (b"TAT", b"TAC", b"TAA"),
                                     }

cdef dict CompUniversalGeneticCode = {b"A" : (b"AGC", b"GGC", b"TGC", b"CGC"),
                                      b"C" : (b"ACA", b"GCA"),
                                      b"D" : (b"ATC", b"GTC"),
                                      b"E" : (b"TTC", b"CTC"),
                                      b"F" : (b"AAA", b"GAA"),
                                      b"G" : (b"ACC", b"GCC", b"TCC", b"CCC"),
                                      b"H" : (b"ATG", b"GTG"),
                                      b"I" : (b"AAT", b"GAT", b"TAT"),
                                      b"K" : (b"TTT", b"CTT"),
                                      b"L" : (b"TAA", b"CAA", b"CTA", b"AAG", b"GAG", b"TAG", b"CAG"),
                                      b"M" : (b"CAT", b"TAT"),
                                      b"N" : (b"ATT", b"GTT", b"TTT"),
                                      b"P" : (b"AGG", b"GGG", b"TGG", b"CGG"),
                                      b"Q" : (b"TTG", b"CTG"),
                                      b"R" : (b"ACG", b"GCG", b"TCG", b"CCG", b"TCT", b"CCT"),
                                      b"S" : (b"AGA", b"GGA", b"TGA", b"CGA", b"ACT", b"GCT", b"TCT", b"CCT"),
                                      b"T" : (b"AGT", b"GGT", b"TGT", b"CGT", b"AAG", b"GAG", b"TAG", b"CAG"),
                                      b"V" : (b"AAC", b"GAC", b"TAC", b"CAC"),
                                      b"W" : (b"TCA", b"CCA"),
                                      b"Y" : (b"ATA", b"GTA", b"TTA"),
                                     }


cdef dict listCodons2Dict(tuple codons):
    cdef dict d={}
    cdef dict s
    cdef dict new
    cdef char* cc
    cdef dict      value
    cdef int  l
    cdef char letter[2]
    cdef bytes c

    letter[1]=0
    
    for c in codons:
        s = d
        cc= c
        for l in range(3):
            letter[0]=cc[l]
            value = s.get(letter,None)
            if value is None:
                value = {}
                s[letter]=value
            s=value
                
    return d

# cpdef dict bindCodons(dict codon1, dict codon2):
#     cdef PyObject* pkey
#     cdef bytes     key
#     cdef PyObject* pvalue  # @DuplicatedSignature
#     cdef dict      l2
#     cdef dict      l3
#     cdef Py_ssize_t ppos1=0
#     cdef Py_ssize_t ppos2=0
#     cdef Py_ssize_t ppos3=0
#     
#     while PyDict_Next(codon1, &ppos1, &pkey, &pvalue):
#         l2=<object>pvalue
#         ppos2=0
#         while PyDict_Next(l2, &ppos2, &pkey, &pvalue):
#             l3=<object>pvalue
#             ppos3=0
#             while PyDict_Next(l3, &ppos3, &pkey, &pvalue):
#                 Py_XDECREF(pvalue)
#                 key=<object>pkey
#                 PyDict_SetItem(l3,key,codon2)
# 
#     return codon1

cpdef dict bindCodons(dict codon1, dict codon2):
    cdef dict      l2
    cdef dict      l3
    
    for l2 in codon1.values():
        for l3 in l2.values():
            for key in l3:
                l3[key]=codon2

    return codon1


cpdef dict codon(char aa, bint direct=True):
    cdef tuple value
    cdef char  letter[2]
    
    letter[0]=aa
    letter[1]=0
    
    
    if direct:
#        value = <object>PyDict_GetItemString(UniversalGeneticCode,letter)
        value = UniversalGeneticCode.get(letter,())
    else:
#        value = <object>PyDict_GetItemString(CompUniversalGeneticCode,letter)        
        value = CompUniversalGeneticCode.get(letter,())        
    
    return listCodons2Dict(value)

cpdef bint homopolymer(bytes s):
    """
    Returns True if the word s is an homopolymer, an homedimer 
    or an homotrimer
    
    @param s: the string to test
    @type s: bytes   
    
    @return True if s is an homopolymer, an homedimer 
            or an homotrimer, False otherwise
    @rtype: bool
    """
    s2 = s[2:] + s[0:2]
    s3 = s[3:] + s[0:3]
    return s==s2 or s==s3
                
def enumerateword(dict automata, bytes prefix=b""):
    cdef bytes k
    cdef bytes w 
    
    if not automata:
        yield prefix
    else:
        for k in automata:
            for w in enumerateword(automata[k],prefix+k):
                yield w
          
cpdef dict word2automata(wordlist):
    cdef dict automata={}
    cdef bytes w
    cdef dict  a 
    cdef bytes l 
    cdef dict  d
    
    for w in wordlist:
        a = automata
        for l in w:
            d = a.get(l,{})
            a[l]=d
            a=d
            
    return automata

cdef double* buildShanonTable(size_t readsize, size_t* offset):
    cdef size_t codonmax = readsize // 3 + (1 if (readsize % 3) else 0)
    cdef size_t i=codonmax+1
    cdef size_t j=0
    cdef size_t k=0
    cdef double* table
    cdef double* line
    cdef double log6=log(6)
    cdef double freq
    
    while i!=0:
        k|=(i & 1) & (i>1)
        i >>= 1
        j+=1
        
    j+=k
    
    offset[0]=j
    codonmax=1<<j
        
    table = <double*> malloc((readsize+1)*(codonmax)*sizeof(double))
    
    for i in range(readsize+1):
        line = table + (i << offset[0])
        for j in range(codonmax):
            if j <=i and j > 0:
                freq = <double> j / <double> i
                freq = -freq * log(freq)/log6
            else:
                freq=0
            line[j]=freq 
            
    return table
            
    
 
cdef bint samereads(unsigned char* p1, unsigned char* p2, size_t length):
    cdef uint64_t *w1 = <uint64_t*>p1
    cdef uint64_t *w2 = <uint64_t*>p2
    cdef size_t   i
    
    for i in range(length):
        if w1[i]!=w2[i]:
            return False
        
    return True   

cdef class AhoCorasick:
    
    def __init__(self,initsize=100000):
        self._finalized = False
        self._step = initsize
        self._depth= 0
        self._maxseqid=0
        self._maxpos=0
        
        # Allocate state arena
        
        self._states = <dnastate*> malloc(sizeof(dnastate)
                                          * initsize)
        
        if self._states==NULL:
            raise RuntimeError("No more memory for automata states")
        
        self._statesize = initsize
        
        # initialize initial state
        
        self._states.match = <protmatch*>-1
        self._states.a     = NULL
        self._states.c     = NULL
        self._states.t     = NULL
        self._states.g     = NULL
        
        self._seqid       = {}
        self._rseqid       = {}
        
        self._statecount = 1
        
        # allocate match arena
        
        self._matches = <protmatch*> malloc(sizeof(protmatch)
                                            * initsize)
        
        if self._matches==NULL:
            raise RuntimeError("No more memory for automata matches")
        
        self._matchsize = initsize
        self._matchcount= 0
        
        
        
    def __dealloc__(self):
        if (self._states != NULL):
            free(<void*>self._states)
        if (self._matches != NULL):
            free(<void*>self._matches)
            
            
    def __len__(self):
        return self._statecount
    
    def addWord(self,bytes word,int protid, int position):
        cdef char* cword = word 
        
        assert not self._finalized, "You cannot add words to a finilized automata"

        self.addCWord(cword,PyBytes_Size(word),protid,position)


    cdef dnastate *getNextState(self,dnastate *base, int letter):
        cdef dnastate** table=&(base.a)
        cdef dnastate*  state = table[letter]
        cdef dnastate*  arena
        cdef size_t     baseid    
    
        if state == NULL:
            if self._finalized:
                return NULL
            else:
                                
                if self._statecount == self._statesize:
                    #
                    # the state arena if full we need to enlarge it
                    #
                    
                    # we save baseid to convert it to the new arena adressing system 
                    baseid = base - self._states
                    
                    # new size of the arena and correponding allocation
                    self._statesize+=self._step
#                    print "realloc buffer to size %d" % self._statesize
                    arena = <dnastate*> realloc(self._states,
                                                sizeof(dnastate)
                                                * self._statesize)
                                         
                    if arena == NULL:
                        # We were not able to enlarge arena
                        return NULL
                    else:
                        self._states = arena
                       
                    #compute the new base and table address
                    base = arena + baseid
                    table=&(base.a)
                
                state = self._states + self._statecount
                table[letter] = <dnastate*>self._statecount
                self._statecount+=1
                
                state.match = <protmatch*>-1
                state.a     = NULL
                state.c     = NULL
                state.t     = NULL
                state.g     = NULL
                
                return state
                
        else:
            if self._finalized:
                return state 
            else:
                return self._states + <size_t> state
            
    cdef protmatch* setAsTerminal(self, dnastate* base, int protid, int position):
        """
        setAsTerminal function is an internal function that is called from the
        addWord method. It must be only used on none finalized automata, but 
        no checking for this property is done a precondition
        """
        cdef protmatch* curmatch = NULL
        cdef protmatch* nextmatch= NULL
        cdef size_t     curmatchid
        cdef protmatch* arena
        
        if position > self._maxpos:
            self._maxpos = position 
        
        if base.match!=<protmatch*>-1:
            nextmatch = self._matches + <size_t>(base.match)        
        
        while(nextmatch):
            curmatch = nextmatch
            if (    curmatch.protid == protid 
                and curmatch.position == position):
                return curmatch
            else:
                if curmatch.next:
                    nextmatch = self._matches + <size_t>curmatch.next
                else:
                    nextmatch = NULL
                    
        # There no match for position@protid register to this state
        # so we have to create a new one
            
        if self._matchcount == self._matchsize:
            # I save the curmatchid to restore curmatch later
            if curmatch :
                curmatchid = curmatch - self._matches
                        
            # the match arena if full we need to enlarge it
            
            self._matchsize+=self._step
#            print "reallocate matches to %d" % self._matchsize
            arena = <protmatch*> realloc(self._matches,
                                         sizeof(protmatch)
                                         * self._matchsize)
                                 
            if arena == NULL:
                # We were not able to enlarge arena
                return NULL
            else:
                self._matches = arena
                
            # I restore curmatch 
            if curmatch :
                curmatch = self._matches + curmatchid

                
        # compute the address 
        nextmatch = self._matches + self._matchcount
        
        if curmatch : 
            curmatch.next = <protmatch*>self._matchcount
        else:
            base.match = <protmatch*>self._matchcount
            
        self._matchcount+=1
        
        nextmatch.protid = protid
        nextmatch.position = position 
        nextmatch.next = NULL 
        
        if protid > self._maxseqid:
            self._maxseqid=protid
        
        return nextmatch 
    
    cdef dnastate* simpleMatch(self,char* cword, size_t lword):
        """
        Checks if cword a dna word of length lword is present in 
        the automata. The function returns the state at the end of
        the word in the automata or null if the word is not found

        This function is used for computing the error link during the
        finalization step
        
        warning : This is a private function that should not be used directly        
        """
        cdef char   letter
        cdef dnastate** table
        cdef dnastate*  curstate
        cdef dnastate*  nextstate
                
        # We cannot match a word of length null
        # neither longer than the longest pattern
        # in the automata
        
        if lword>self._depth:
            return NULL 
        
        nextstate = self._states 
        
        while (nextstate!=NULL and lword):
            curstate = nextstate

            letter = cword[0]
            table    = &(curstate.a)
            nextstate = table[letter]
            
            cword+=1
            lword-=1
            
        if lword==0 and nextstate!=NULL:
            return nextstate
        else:
            return NULL
        
        
    cdef dnastate* longestSuffix(self,char* cword, size_t lword):
        """
        identify the longest suffix of cword that is recognized by
        the Aho-Corasick automata.
        
        This function is used for computing the error link during the
        finalization step
        
        warning : This is a private function that should not be used directly        
        """
        cdef dnastate* suffix=NULL

        
        while lword >= 1 and suffix==NULL:
            cword+=1
            lword-=1

            suffix = self.simpleMatch(cword,lword)
            
        return suffix
            
        
    cdef setErrorLink(self):
        """
        Add error link to the Aho-Corasick automata during its finalization
        step.
        
        warning : This is a private function that should not be used directly
        """
        cdef char* buffer
        cdef size_t*     stack
        cdef dnastate*  curstate
        cdef dnastate*  nextstate
        cdef dnastate** table
        cdef dnastate* links
        cdef dnastate** errors
        cdef size_t i
        cdef size_t j
        cdef int pos=0
        cdef size_t lsuffix
        cdef size_t lword
         

#        print "depth : %d -> %d" % (self._depth,sizeof(size_t)*(self._depth+1)*4)

        stack = <size_t*>malloc(sizeof(size_t)*(self._depth+1)*4)

        errors = <dnastate**>malloc(sizeof(dnastate*) * self._statecount)
        bzero(<void*>errors,sizeof(dnastate*) * self._statecount)
        
        buffer=<char*>malloc(sizeof(char)*(self._depth+1))

        #
        # We explore trie with a deep first algorithm
        #
        
        # I put on the stack the root of the automata
        stack[0] = 0
        pos = 0

        j=0 # used to scan the four branches of a node     
       

        while pos >= 0:
#            print "Run on state : [%d]  %d-%d" % (pos,stack[pos],j)

            # I consider the last state on the stack 
            curstate = self._states + stack[pos] 
            table=&(curstate.a)          
            
            # We skip the empty link
            while j < 4 and table[j]==NULL:
                j+=1
                
            # on the loop exit or j==4 and we have explored all the branches
            # or j<4 and we found a non empty branch
                
            if j < 4:
                
                # We found a link we push it on the stack
                
                buffer[pos]=j
                pos+=1
                stack[pos]= (table[j] - self._states)
#                print "push %d [%d]" % (pos,j)
                j=0
                            
                # ===> Then we loop on the main while
                
            else:
                
                # There is no more link on this state
                # --> we cannot go deeper
                                
                links = self.longestSuffix(buffer,pos)
                errors[stack[pos]]=links
                                           
                # we pop out the state
                pos-=1
                j = buffer[pos]+1
         
        errors[0]=NULL       

#        for i in range(1,self._statecount):
#            if errors[i]==NULL:
#                print "no link on %d" % i
                
        for i in range(self._statecount):
            curstate=self._states + i 
            for j in range(4):
                nextstate=curstate
                table=&(nextstate.a)
                if table[j]==NULL:
                    while nextstate!=NULL and table[j]==NULL:
                        nextstate = errors[nextstate - self._states]
                        table=&(nextstate.a)
                    if nextstate==NULL:
#                        print "jump to %d[%d] -> root" % (i,j)
                        nextstate = self._states
                    else:
                        nextstate = table[j]
#                        print "jump to %d[%d] -> %d" % (i,j,nextstate-self._states)
                    table =&(curstate.a) 
                    table[j]=nextstate
                
                
        free(errors)
        free(stack)
        free(buffer)

    cpdef finalize(self):
        cdef size_t i 
        cdef size_t j 
        cdef size_t dest
        cdef dnastate* state
        cdef dnastate** table  # @DuplicatedSignature
        cdef protmatch* match
        
        if self._finalized:
            return None
        
        self._finalized = True

        for i in range(self._statecount):
            state = self._states + i
            
            if state.match!=<protmatch*>-1:
                state.match = self._matches + <size_t>state.match
            else:
                state.match = NULL
                
            table=&(state.a)
            for j in range(4):
                if table[j]!=NULL:
                    table[j]=self._states + <size_t>table[j]
                    
        for i in range(self._matchcount):
            match = self._matches + i
            if match.next!=NULL:
                match.next=self._matches + <size_t>match.next
                                                            
        self.setErrorLink()          
        
      
    cdef int addCWord(self,char* cword, size_t lword, int protid, int position):
        cdef int   letter
        cdef int   i
        cdef dnastate* state = self._states
        cdef protmatch* match  # @DuplicatedSignature
                        
        for i in range(lword):
            letter = cword[i] 
            letter >>=1
            letter &=3
            state = self.getNextState(state,letter)
            
            if state==NULL:
                return 0
            
        match = self.setAsTerminal(state,protid,position)
            
        if match == NULL:
            return 0
         
        if lword > self._depth:
            self._depth=lword  
             
        return self._statecount
    
    
    cdef int addAutomata(self,dict automata,int protid, int position):
        cdef list       stack=[(automata,0,0)]
        cdef tuple      stackitem
        cdef PyObject*  pkey
        cdef char*      ckey  
        cdef int        letter
        cdef PyObject*  pvalue
        cdef dict       value
        cdef Py_ssize_t ppos=0
        cdef dnastate*  state
        cdef dnastate*  nstate
        cdef protmatch* match
        cdef size_t     stateid
        cdef size_t     lword=0
       
        while PyList_GET_SIZE(stack)>0 :
            stackitem = stack.pop()
            automata  = <object>PyTuple_GET_ITEM(stackitem,0)
            stateid   = PyInt_AsSsize_t(<object>PyTuple_GET_ITEM(stackitem,1))
            lword     = PyInt_AsSsize_t(<object>PyTuple_GET_ITEM(stackitem,2))
            state     = self._states + stateid
            ppos      = 0
            
            lword+=1
            
            while PyDict_Next(automata, &ppos, &pkey, &pvalue):
                
                if lword > self._depth:
                    self._depth=lword
                #
                # Get the nucleotide code
                #
                
                ckey = PyBytes_AS_STRING(pkey)
                letter = ckey[0]
                letter>>=1
                letter&=3
                
                #
                # create or gate the corresponding node in the automata
                #
                
                nstate = self.getNextState(state,letter)
                
                # potentially recompute state address if reallocation occured 
                state     = self._states + stateid
                
                if state==NULL:
                    return 0
                
                # 
                # if value is an empty dict then this is a terminal state
                # else we have to push it on the stack
                #
                
                value=<object>pvalue
                
#                print stateid,"->",(nstate - self._states),value
                
                if PyDict_Size(value):
                    stack.append((value,(nstate - self._states),lword))
#                    print "on continue"
                else:
#                    print "Etat final"
                    match = self.setAsTerminal(nstate,protid,position)       
#                    print "Etat final ok"
                    if match == NULL:
                        return 0
                   
        return self._statecount
                
        
    
    #cython: initializedcheck=False 
    cpdef list match(self,char* sequence):
#        cdef char*  csequence = sequence 
        cdef size_t lseq      = len(sequence)
        cdef size_t pos      
        cdef dnastate* state 
        cdef dnastate* nstate 
        cdef dnastate** table 
        cdef protmatch* pmatch
        cdef int letter
        cdef int rsize = self._maxseqid * 2 + 1
#        cdef list results = PyList_New(rsize)
        cdef list results = [[] for i in range(rsize)]
        cdef list lpos
        cdef int protid
        
#        for i in range(rsize):
#            PyList_SET_ITEM(results,i,[])
                
        if not self._finalized:
            self.finalize()
        
        state = self._states
        
        for pos in range(lseq):
            table = &(state.a)
            letter = sequence[pos]
            letter >>=1
            letter &=3
            nstate = table[letter]
            pmatch = nstate.match
            while (pmatch!=NULL):
                if pmatch.protid < 0:
                    lpos = <object>PyList_GET_ITEM(results,rsize + pmatch.protid)
                    #lpos=results[rsize + pmatch.protid]
                else:
                    lpos = <object>PyList_GET_ITEM(results,pmatch.protid)
                    #lpos=results[pmatch.protid]
                    
                PyList_Append(lpos,(pos,pmatch.position))
                #lpos.append((pos,pmatch.position))
                pmatch = pmatch.next
            state = nstate
                    
        return results 
    
    cpdef DiGraphMultiEdge asGraph(self):
        cdef DiGraphMultiEdge g 
        cdef size_t i 
        cdef size_t j 
        cdef size_t dest
        cdef dnastate* ndest
        cdef dict node
        cdef dict edge
        cdef dnastate** table
        cdef protmatch* match  # @DuplicatedSignature
        cdef size_t matchcount = 0
        cdef size_t pnode
        
        g=DiGraphMultiEdge(b"AhoCorasick")
        
        for i in range(self._statecount):
            g.addNode(i)
            node=g.getNodeAttr(i)
            node[b'label'] = b"%d" % i
            node[b'graphics'] = {}
            if i==0:
                node[b'graphics'][b'fill']=b'#008000'
                node[b'graphics'][b'type']=b'rectangle'  
                
            match = self._states[i].match    
            if ((not self._finalized and match!=<protmatch*>-1) or
                (self._finalized and match!=NULL)):
                node[b'graphics'][b'fill']=b'#800000'
                node[b'graphics'][b'type']=b'triangle'
                
                if not self._finalized:
                    match = self._matches + <size_t>match
                
                pnode = i
                while(match!=NULL):
                    #print "%d" % <size_t>match
                    matchcount+=1
                    g.addNode(self._statecount + matchcount)
                    node = g.getNodeAttr(self._statecount + matchcount)
                    node[b'graphics'] = {"type" : 'circle',
                                         'fill' : "#000080"
                                        }
                    node[b'label'] = "%d,%d" % (match.protid,match.position)
                    g.addEdge(pnode,self._statecount + matchcount)  
                    edge = g.getEdgeAttr(pnode,self._statecount + matchcount)
                    edge['graphics']={b'arrow':b'last',
                                      b'fill':b'#000040'
                                     }            
                    pnode = self._statecount + matchcount
                    match = match.next
                    if match!=NULL and not self._finalized:
                        match = self._matches + <size_t>match
                
        for i in range(self._statecount):
            table=&(self._states[i].a)
            for j in range(4):
                if self._finalized:
                    ndest = table[j]
                    if ndest:
                        dest  = ndest - self._states
                    else:
                        dest = 0
                else:
                    dest  = <size_t>table[j]
                    ndest = <dnastate*>-1
                    if dest==0:
                        ndest=NULL
                    
                if ndest != NULL:
                    g.addEdge(i,dest)
                    edge = g.getEdgeAttr(i,dest)
                    edge[b"label"]=b'%s' % (b"actg"[j])
                    edge['graphics']={b'arrow':b'last',
                                      b'fill':b'#080000'
                                     }            
        
        return g
    

cdef class ProtAhoCorasick(AhoCorasick):

    cpdef int addSequence(self,bytes sequence, object seqid=1, size_t kup=4):
        cdef int    position
        cdef int    k
        cdef char*  csequence
        cdef char   letter[2]
        cdef dict   c
        cdef dict   cn
        cdef int    osize
        cdef size_t erreur
        cdef int    id
        
        if self._finalized:
            return 0
        
        
        if seqid in self._seqid:
            id = self._seqid[seqid]
        else:
            id = self._maxseqid + 1
            self._seqid[seqid]=id
            self._rseqid[id]=seqid
            
        # register the size of the automata before add the protein
                
        osize = self._statecount
        
        letter[1]=0
        csequence = sequence
                
        # Backtranslate according to the forward strand
        
        for position in range(len(sequence)-kup+1):
            
            # for each peptide of size kup build a small automata with all
            # codon variants
            
            c = codon(csequence[position+kup-1])
            
            if c:
                erreur=0
                for k in range(position+kup-2,position-1,-1):
                    cn = codon(csequence[k])
                    if cn:
                        c = bindCodons(cn,c)
                                                    
                    else:
                        erreur=1
                        
                if erreur==0:
                    c = word2automata([w for w in enumerateword(c) if not homopolymer(w)])
                    erreur = self.addAutomata(c, id, (position+kup) * 3 - 1)
                    
                    if erreur==0:
                        return 0
                                
        # Backtranslate according to the reverse strand
                
        for position in range(len(sequence)-1,kup-2,-1):
            
            # for each peptide of size kup build a small automata with all
            # anticodon variants
            
            c = codon(csequence[position-kup+1],False)
            
            if c:
                erreur=0
                for k in range(position-kup+2,position+1):
                    cn = codon(csequence[k],False)
                    if cn:
                        c = bindCodons(cn,c)
                    else:
                        erreur=1
                        
                if erreur==0:
                    c = word2automata([w for w in enumerateword(c) if not homopolymer(w)])
        
                    erreur = self.addAutomata(c, -id, position * 3 + 2)
                    if erreur==0:
                        return 0

        return self._statecount - osize

    cpdef scanIndex(self,Index seqindex,int minmatch=15, int maxmatch=-1, int covmin=1, bint progress=True):
        """
        
        
            @param seqindex: A NGS read index
            @type seqindex: orgasm.indexer._orgasm.Index instance
            
            @param minmatch: minimum count of identified word to take into
                             account the match
            @type minimum: int
                             
            @param maxmatch: maximum count of identified word to take into
                             account the match. Repeated can leads to huge
                             wordcount without real similarity
                             By default (-1) this value is set to readsize/3
                             (I.e amino acid count).
            @type maximum: int
            
            @param covmin: minimum count observed for a read to be taken into 
                           account
            @type covmin: int
            
            @param progress: Boolean flg indicating if a progress bar must be
                             displayed
            @type progress: bool
        """
        cdef unsigned char* nucs
        cdef size_t i  # @DuplicatedSignature
        cdef size_t j  # @DuplicatedSignature
        cdef int k
        cdef list ks
        cdef size_t readcount  = seqindex._index.readCount
        cdef size_t recordsize = seqindex._index.recordSize
        cdef size_t readsize   = seqindex._index.readSize
        
        # the number of non-overlaping 4mer in a read
        cdef size_t readsize4  = (readsize >> 2) + (1 if (readsize & 3)  else 0)
        
        # the number of non-overlaping 32mer in a read
        cdef size_t readsize64 = (readsize >> 5) + (1 if (readsize & 31) else 0)
        
        # Pointer to the first read record
        cdef unsigned char*  records    = <unsigned char*>seqindex._index.records
        cdef unsigned char*  nuc4
        cdef unsigned char*  nuc4bis
        cdef BitVector   wordcount
        cdef size_t wordmax=0
        cdef double shamin
        cdef size_t wc[6]
        cdef size_t rp[6]
        cdef dict matchpos
        
        cdef size_t* count
        cdef int wct
        cdef int wcmax
        cdef int phase 
        cdef int protidmax=0
        cdef int framemax=0
        cdef int loc
        cdef int locmax
        
        cdef size_t nbreads
        cdef dnastate* state 
        cdef dnastate* nstate 
        cdef dnastate** table 
        cdef int letter
        cdef dict results = {}
        cdef int readid
        cdef PyObject*  plpos        
        cdef list lpos
        cdef int pid 
        cdef int p 
        cdef double  shanon
        cdef size_t  offset
        cdef double* shanonTable = buildShanonTable(readsize,&offset)
        cdef double* shanonLine
        cdef ProgressBar pb
        
        # mid is the number of read frame globally available for each 
        # protein stored in the automata
        cdef int mid = self._maxseqid * 6 + 3
        
        # Finalize the automata if needed
        if not self._finalized:
            self.finalize()
        
        # Computes the default maxmatch according to the read size
        if maxmatch < 0:
            maxmatch = readsize // 3
            
        i=0
        
        wordcount = BitVector(mid,self._maxpos)
        
        nuc4 = records + i * (recordsize)
        
        if (progress):
            pb=ProgressBar(readcount)
#            progressBar(1,readcount,True)

        while i < readcount:
                
            # We start a new read so we set counter to 0
            wordcount.clear()     
            matchpos={}   
            
            readid = i
#            print "@",i
            
            # setup automata to the initial state
            state = self._states
            table = &(state.a)
            for j in range(readsize4):
                nucs = <unsigned char*>&(expanded8bitsnuc[nuc4[j]])
                for k in range(4):
                    letter = nucs[k]
                    nstate = table[letter]
                    pmatch = nstate.match
                    while (pmatch!=NULL):
#                        print "coucou %d %d" %(pmatch.protid,pmatch.position)
                        pid = pmatch.protid * 3 + ((j*4+k) % 3)
                        if pid < 0 :
                            pid += mid
                        wordcount.set(pid,pmatch.position)
                        matchpos[pid]=min(matchpos.get(pid,65535),pmatch.position)
                        pmatch = pmatch.next
                    state = nstate
                    table = &(state.a)
                    
            shamin=1.0
            wordmax=0
            protidmax=0
            framemax=0
            
            count = wordcount.counter()
            
            for k in range(1,self._maxseqid+1):
                pid = k * 3
                
                wc[3]=count[pid]
                wc[4]=count[pid+1]
                wc[5]=count[pid+2]
                
                rp[3]=matchpos.get(pid,65535)
                rp[4]=matchpos.get(pid+1,65535)
                rp[5]=matchpos.get(pid+2,65535)
                
                #memcpy(<void*>(wc+3),<void*>(count+pid),3*sizeof(size_t))
                
                pid = mid - pid 
                 
                wc[0]=count[pid]
                wc[1]=count[pid+1]
                wc[2]=count[pid+2]

                rp[0]=matchpos.get(pid,65535)
                rp[1]=matchpos.get(pid+1,65535)
                rp[2]=matchpos.get(pid+2,65535)
                
                # memcpy(<void*>(wc),<void*>(count+pid),3*sizeof(size_t))
                
                wct = wc[0] + wc[1] + wc[2] + \
                      wc[3] + wc[4] + wc[5]
                                            
                if wct > minmatch:
                    shanonLine = shanonTable + (wct << offset)
                    shanon=shanonLine[wc[0]] + \
                           shanonLine[wc[1]] + \
                           shanonLine[wc[2]] + \
                           shanonLine[wc[3]] + \
                           shanonLine[wc[4]] + \
                           shanonLine[wc[5]]
                
                    wcmax=0
                    phase=0
                    for p in range(6):
                        if wc[p]>wcmax:
                            wcmax=wc[p]
                            phase=p-3
                            loc=rp[p]
                else:
                    shanon=1.0
                    wcmax=0
                    phase=0

                if ((shanon<0.1) and
                    (wcmax >=minmatch) and
                    (wcmax >wordmax)):
                    shamin = shanon
                    wordmax = wcmax
                    protidmax=k
                    framemax= phase
                    locmax=loc
                    

                
            nbreads=1
            nuc4bis=nuc4 + recordsize
            i+=1

            if (progress):
                pb(i)

            while (i< readcount and 
                   samereads(nuc4, nuc4bis, readsize64)):
                nbreads+=1
                nuc4bis+= recordsize
                i+=1
                
#            print "Best match : %d -> %d (%d)" % (protidmax,wordmax,nbreads) 

            nuc4 = nuc4bis
            
            if protidmax>0 and shamin < 0.1 and nbreads >= covmin:
                if framemax < 0:
                    readid = -readid - 1
                else:
                    readid+=1 
                    
                plpos = PyDict_GetItem(results,PyInt_FromLong(protidmax))
                if plpos==NULL:
                    lpos = []
                    PyDict_SetItem(results,PyInt_FromLong(protidmax),lpos)
                else:
                    lpos = <object>plpos
                lpos.append((readid,wordmax,nbreads,framemax,shamin,locmax))
        
        #<------------------------------ End of the while loop ------------------------>
                
        ks = list(results.keys())

        for k in ks:
            results[self._rseqid[k]]=results[k]
            del results[k]
        

        free(shanonTable)          
        return results
                
 
cdef class NucAhoCorasick(AhoCorasick):

    cpdef int addSequence(self,bytes sequence, object seqid=1, size_t kup=12):
        cdef int   position
        cdef int   k
        cdef char* csequence
        cdef char  letter[2]
        cdef dict  c
        cdef int osize
        cdef size_t erreur
        cdef int   id
        
        if self._finalized:
            return 0
        
        
        if self._seqid.has_key(seqid):
            id = self._seqid[seqid]
        else:
            id = self._maxseqid + 1
            self._seqid[seqid]=id
            self._rseqid[id]=seqid
            
        # register the size of the automata before add the protein
        
        osize = self._statecount
        
        # Backtranslate according to the forward strand

        for position in range(len(sequence)-kup+1):
            
            # for each peptide of size kup build a small automata with all
            # codon variants
            self.addWord(sequence[position:position+kup],id,position)
            self.addWord(reverseComplement(sequence[position:position+kup]),-id,position)

        return self._statecount - osize
    
    cpdef scanIndex(self,Index seqindex,int minmatch=5, int maxmatch=-1, int covmin=1, bint progress=True):
        """
        
        
            @param seqindex: A NGS read index
            @type seqindex: orgasm.indexer._orgasm.Index instance
            
            @param minmatch: minimum count of identified word to take into
                             account the match
            @type minimum: int
                             
            @param maxmatch: maximum count of identified word to take into
                             account the match. Repeated can leads to huge
                             wordcount without real similarity
                             By default (-1) this value is set to readsize/3
                             (I.e amino acid count).
            @type maximum: int
            
            @param covmin: minimum count observed for a read to be taken into 
                           account
            @type covmin: int
            
            @param progress: Boolean flg indicating if a progress bar must be
                             displayed
            @type progress: bool
        """
        cdef unsigned char* nucs
        cdef size_t i  # @DuplicatedSignature
        cdef size_t j  # @DuplicatedSignature
        cdef int k
        cdef list ks
        cdef size_t readcount  = seqindex._index.readCount
        cdef size_t recordsize = seqindex._index.recordSize
        cdef size_t readsize   = seqindex._index.readSize
        
        # the number of non-overlaping 4mer in a read
        cdef size_t readsize4  = (readsize >> 2) + (1 if (readsize & 3)  else 0)
        
        # the number of non-overlaping 32mer in a read
        cdef size_t readsize64 = (readsize >> 5) + (1 if (readsize & 31) else 0)
        
        # Pointer to the first read record
        cdef unsigned char*  records    = <unsigned char*>seqindex._index.records
        cdef unsigned char*  nuc4
        cdef unsigned char*  nuc4bis
        cdef BitVector   wordcount
        cdef size_t wordmax=0
        cdef double shamin
        cdef size_t wc[6]
        cdef size_t rp[6]
        cdef dict matchpos
        
        cdef size_t* count
        cdef int wct
        cdef int wcmax
        cdef int phase 
        cdef int protidmax=0
        cdef int framemax=0
        cdef int loc
        cdef int locmax
        
        cdef size_t nbreads
        cdef dnastate* state 
        cdef dnastate* nstate 
        cdef dnastate** table 
        cdef int letter
        cdef dict results = {}
        cdef int readid
        cdef PyObject*  plpos        
        cdef list lpos
        cdef int pid 
        cdef int p 
        cdef size_t  offset
        
        # mid is the number of read frame globally available for each 
        # protein stored in the automata
        cdef int mid = self._maxseqid * 2 + 1
        # Finalize the automata if needed
        if not self._finalized:
            self.finalize()
        
        # Computes the default maxmatch according to the read size
        if maxmatch < 0:
            maxmatch = readsize 
            
        i=0
        
        wordcount = BitVector(mid,self._maxpos)
        
        nuc4 = records + i * (recordsize)
        
        if (progress):
            pb=ProgressBar(readcount)

        while i < readcount:
                
            # We start a new read so we set counter to 0
            wordcount.clear()     
            matchpos={}   
            
            readid = i
#            print "@",i
            
            # setup automata to the initial state
            state = self._states
            table = &(state.a)
            for j in range(readsize4):
                nucs = <unsigned char*>&(expanded8bitsnuc[nuc4[j]])
                for k in range(4):
                    letter = nucs[k]
                    nstate = table[letter]
                    pmatch = nstate.match
                    while (pmatch!=NULL):
                        pid = pmatch.protid
                        if pid < 0 :
                            pid += mid
                        wordcount.set(pid,pmatch.position)
                        matchpos[pid]=min(matchpos.get(pid,65535),pmatch.position)
                        pmatch = pmatch.next
                    state = nstate
                    table = &(state.a)
                    
            wordmax=0
            protidmax=0
            framemax=0
            
            count = wordcount.counter()
            
            for k in range(1,self._maxseqid+1):
                pid = k 
                
                wc[1]=count[pid]
                rp[1]=matchpos.get(pid,65535)                
                #memcpy(<void*>(wc+3),<void*>(count+pid),3*sizeof(size_t))
                
                pid = mid - pid 
                 
                wc[0]=count[pid]
                rp[0]=matchpos.get(pid,65535)
                
                # memcpy(<void*>(wc),<void*>(count+pid),3*sizeof(size_t))
                
                wct = wc[0] + wc[1]
                if wc[0] > wc[1]:
                    phase=-1
                    wcmax=wc[0]
                    loc=rp[0]
                else:
                    phase=1
                    wcmax=wc[1]
                    loc=rp[1]
                

                if (wcmax >=minmatch):
                    wordmax = wcmax
                    protidmax=k
                    framemax= phase 
                    locmax=loc

                
            nbreads=1
            nuc4bis=nuc4 + recordsize
            i+=1

            if (progress):
                pb(i)

            while (i< readcount and 
                   samereads(nuc4, nuc4bis, readsize64)):
                nbreads+=1
                nuc4bis+= recordsize
                i+=1
                
#            print "Best match : %d -> %d (%d)" % (protidmax,wordmax,nbreads) 

            nuc4 = nuc4bis
            
            if protidmax>0 and nbreads >= covmin:
                if framemax < 0:
                    readid = -readid - 1
                else:
                    readid+=1 
                    
                plpos = PyDict_GetItem(results,PyInt_FromLong(protidmax))
                if plpos==NULL:
                    lpos = []
                    PyDict_SetItem(results,PyInt_FromLong(protidmax),lpos)
                else:
                    lpos = <object>plpos
                lpos.append((readid,wordmax,nbreads,framemax,0,locmax))
                
                
        ks = list(results.keys())

        for k in ks:
            results[self._rseqid[k]]=results[k]
            del results[k]
        

        return results
                    


# cython: language_level=3

cdef extern from "stdio.h":
    struct FILE
    int fprintf(FILE *stream, char *format, ...)
    FILE* stderr
    ctypedef unsigned int off_t "unsigned long long"
    
    

cdef extern from "time.h":
    struct tm :
        int tm_yday 
        int tm_hour
        int tm_min
        int tm_sec

    enum: CLOCKS_PER_SEC
    
    ctypedef int time_t
    ctypedef int clock_t
    
    
    tm *gmtime_r(time_t *clock, tm *result)
    time_t time(time_t *tloc)
    clock_t clock()

cpdef object progressBar(object pos,
                         off_t maxi,
                         bint reset=?,
                         bytes head=?,
                         list delta=?,
                         list step=?)

    


'''
Created on 13 mars 2013

@author: coissac
'''



def fasta(filename):
    
    data = open(filename).readlines()
    prots = {}
    protid= None
    seq = None
        
    for l in data:
        if l[0]=='>':
            protid = l.split()[0][1:]
            seq = prots.get(protid,[])
            prots[protid]=seq
        else:
            seq.append(l.strip())
            
    for p in prots:
        prots[p]=bytes("".join(prots[p]).upper(),encoding="ascii")
        
    return prots

#
#def fasta(filename,kmer=4):
#    
#    data = open(filename).readlines()
#    prots = {}
#    protid= None
#    seq = None
#    
#    hmer = set(i * 4 for i in 'ACDEFGHIKLMNOPQRSTVWY')
#        
#    for l in data:
#        if l[0]=='>':
#            if seq is not None:
#                seq = "".join(seq).upper()
#                seq = set(seq[i:(i+kmer)] for i in range(len(seq)-kmer))
#                prots[protid]=prots.get(protid,set()) | seq
#            protid = l.split()[0][1:]
#            seq = []
#        else:
#            seq.append(l.strip())
#            
#    if seq:
#        seq = "".join(seq).upper()
#        seq = set(seq[i:(i+kmer)] for i in range(len(seq)-kmer))
#        prots[protid]=prots.get(protid,set()) | seq
#        
#    kmers = {}
#    
#    for protid in prots:
#        prots[protid]-=hmer
#        for k in prots[protid]:
#            kmers[k]=kmers.get(k,set()) | set([protid])
#            
#    kmers = dict((k,kmers[k]) for k in kmers if len(kmers[k]) < 10)
#        
#    pk={}
#    for k,ps in kmers.iteritems():
#        protid = ps.pop()
#        pk[protid] = pk.get(protid,set()) | set([k])
#        
#        
#    return prots,kmers,pk
#        
#        
        
        


# cython: language_level=3


from orgasm.indexer._orgasm cimport Index
from ._asmbgraph cimport AsmbGraph
from ._assembler cimport Assembler
from cpython.list cimport PyList_GET_SIZE
from functools import reduce
import sys

# cdef int isanchors(Index index ,int sid):
#     return index.getIds(index.getPairedRead(sid))[0]


cdef set setofreads(dict extensions):
    cdef set back=set()
    
    for x in extensions.values():
        back|=set(x[1])
        
    return back

cdef float isgrowing(list lastseeds):
    cdef int llast = PyList_GET_SIZE(lastseeds)
    cdef double i
    cdef int p = 0
    
    for i in lastseeds:
        if   i > 1.:
            p+=1
        elif i < 1.:
            p-=1
            
    return p / llast


cdef set __used_reads__ = set()

cpdef set getusedreads():
    return __used_reads__

cpdef resetusedreads():
    __used_reads__.clear()

cpdef AsmbGraph tango(Assembler self,
          list seeds,
          int minread=10,    float minratio=0.1, 
          int mincov=1,      int   minoverlap=40,
          bint lowfilter=True, 
          tuple adapters5=(), 
          tuple adapters3=(), 
          maxjump=0,restrict=None,
          int cycle=1, int nodeLimit=1000000,
          bint progress=True,
          bint useonce=True,
          logger=None):
    '''
    the :py:func:`~organsm.assembler.tango` function is the main assembling function. It extends selected
    seeds to produce the assembled sequence.
    
    :param seeds: a :py:class:`list` describing the extensions points. Each element of the list corresponds to 
                  a seed. A seed is described by a :py:class:`tuple` of two elements. 
                  
                  The first element is an integer value corresponding to the ``id`` of a read. if the ``id`` is greater than 0
                  the seed will be extended only if it is not already included in the assembling graph.
                  If the ``id`` is specified as a negative number, the seed will be included in the procedure
                  extension whatever if it is already member or not of the assembling graph.
                  
                  The second element of the :py:class:`tuple` is itself a :py:class:`tuple` contenaing or the 
                  :py:class:`None` value of a :py:class:`str` elements used to annotate the read.
                  
                  .. code-block :: python
                  
                      [(12347, ('matK',),0),(-37247,(None,),0)]
                  
                  In this example the structure describe two seeds. The first one corresponds to the
                  read 12347 annotated as belonging to the *matK* gene. The second corresponds to the
                  read 37247, without annotation. The process of extension will be run for the first seed
                  only if it is not already member of the assembling graph and always run on the second seed.
                  
    :type seeds: a :py:class:`list` of :py:class:`tuple`.
                     
    :param minread: Minimum count of reads having to be taken into account to extend an assembling path.
                    If this minimum count is not reach the extension of this path is stopped
    :type minread: :py:class:`int`
    
    :param minratio: A branch is added to the current path if the corresponding symbole **A**, **C**, **G** or **T**
                     represents at least a fraction of ``minratio``of the observed reads.
    :type minratio: :py:class:`float`
    
    :param mincov: To be considered, a read must be observed at least ``mincov`` in the total data set.
    :type mincov: :py:class:`int`
    
    :param minoverlap: Only reads overlaping the current extension point by ``minoverlap`` base pairs are
                       considered for the extension process. 
    :type minoverlap: :py:class:`int`
    
    :param lowfilter:
    :type lowfilter: :py:class:`bool`
    :param maxjump:
    :type maxjump: :py:class:`int`
    :param restrict:
    :type restrict: :py:class:`set` of :py:class:`int` values
    '''
        
    cdef int delta= 0
    cdef int sdelta= 0
    cdef str wheel= '|/-\\'
    cdef int lastprint=-1
    
    # a direct pointer to the read index
    cdef Index index = self._index
    
    # a direct pointer to the graph structure
    cdef AsmbGraph graph = self._graph
    
    # a direct access to the read length
    cdef int readLength = index.getReadSize()
    cdef int readmax    = index.len() + 1
    
    # extension cycle counter for the progress status monitoring
    cdef int icycle=0
    
    # count of missing reads in the graph (aka fake reads)
    cdef int fake=0
    
    # list of the stack size during the last 1000 cycles 
#    lastcycle=[]
    cdef list lastseeds=[0]
    cdef int addedseed=0
    
    # count of pair end jump on an extension branch
    # maxanchor = 0
    
    cdef str lastgene=None
    cdef float growing
    cdef int readid
    cdef int lgraph
    
    cdef dict rextensions
    cdef dict lextensions
    cdef dict backe
    cdef set back
    cdef list rk
    
    assert seeds is not None,'seeds cannot be set to None'
    
    while (PyList_GET_SIZE(seeds) > 0) and (graph.nodeCount() < nodeLimit) :
        
        #time.sleep(1)
        icycle+=1
        
        # Counter used to check how many seed will be added 
        # to the stack during the cycle
        addedseed=0

        # We pop one seed from the stack
        
        readid,(gene,),sdelta = seeds.pop()
        
        if useonce:
            __used_reads__.add(readid)
            
        if sdelta < delta:
            delta = sdelta
            
        if gene is not None:
            lastgene = gene
        
#         # We test if the read poped out from the stack
#         # is just a jump mark
#         if readid==0:
#             maxanchor-=1
#             continue
        
        # Extension is forced if readid is provided as a negative number
        if readid < 0 or readid not in graph:
            
            if readid < 0:
                readid=-readid
            #print "\n==>",readid,str(gene)
            
            growing = isgrowing(lastseeds)
#             growing = (  float(sum([ls > 1. for ls in lastseeds])) \
#                        - float(sum([ls < 1. for ls in lastseeds]))) / len(lastseeds)
            
            if growing < -0.01 and delta > 0:
                delta-=1
                lastseeds=[1]*1000
            if growing > +0.01 and delta < 40:
                delta+=1
                lastseeds=[1]*1000
            
            if progress and (icycle % 50)==0 and sys.stderr.isatty(): 
                print("%s : %d bp [%4.1f%% fake reads; Stack size: %8d / %6.2f %d  Gene: %s " % (wheel[(icycle//10) % 4],
                                                                                                 len(graph)/2,
                                                                                                 float(fake)/(len(graph)+1)*200.,
                                                                                                 len(seeds),
                                                                                                 growing,
                                                                                                 delta,
                                                                                                 lastgene),
                       file=sys.stderr,
                       end="\r")
                      
            lgraph = graph.nodeCount()                                                                                           
            if logger is not None and lgraph!=lastprint and lgraph % 20000==0:
                lastprint=lgraph
                logger.info("%d bp [%4.1f%% fake reads; Stack size: %8d / %6.2f %d  Gene: %s " % (lgraph//2,
                                                                                                  float(fake)/(lgraph+1)*200.,
                                                                                                   PyList_GET_SIZE(seeds),
                                                                                                   growing,
                                                                                                   delta,
                                                                                                   lastgene))
                
    
            #
            # We add the read into the graph
            #
                
            #
            # extends the read on the right side
            #
            #logger.info("Before RCheck %s" % index.getRead(readid,1,readLength-1))
            rextensions = index.checkedExtensions(probe      = index.getRead(readid,1,readLength-1), 
                                                  adapters5  = adapters5,
                                                  adapters3  = adapters3,
                                                  minread    = minread+delta, 
                                                  minratio   = minratio, 
                                                  mincov     = mincov,  
                                                  minoverlap = minoverlap+delta, 
                                                  extlength  = 1,
                                                  lowfilter  = lowfilter,
                                                  restrict   = restrict,
                                                  exact      = True)
            #print >>sys.stderr,"After RCheck"
            #print(rextensions)
            #
            # Build a bijective relationship for the right extensions
            #
            rk=list(rextensions.keys())
            for k in rk:
                #print >>sys.stderr,"Before RICheck",index.getRead(-rextensions[k][1][0],1,readLength-1)
                backe = index.checkedExtensions(probe      = index.getRead(-rextensions[k][1][0],1,readLength-1), 
                                                adapters5  = adapters5,
                                                adapters3  = adapters3,
                                                minread    = minread+delta, 
                                                minratio   = minratio, 
                                                mincov     = mincov,  
                                                minoverlap = minoverlap+delta, 
                                                extlength  = 1,
                                                lowfilter  = lowfilter,
                                                restrict   = restrict,
                                                exact      = True)
                #print >>sys.stderr,"After RICheck"

                # back = set(reduce(lambda p,q:p+q,[backe[y][1] for y in backe],[]))
                back = setofreads(backe)
                
                if -readid not in back:
                    #print "Remove R extension : ",k
                    del rextensions[k]
            
    
            #print >>sys.stderr,"Before LCheck",index.getRead(-readid,1,readLength-1)
            lextensions = index.checkedExtensions(probe      = index.getRead(-readid,1,readLength-1), 
                                                  adapters5   = adapters5,
                                                  adapters3   = adapters3,
                                                  minread    = minread+delta, 
                                                  minratio   = minratio, 
                                                  mincov     = mincov,  
                                                  minoverlap = minoverlap+delta, 
                                                  extlength  = 1,
                                                  lowfilter  = lowfilter,
                                                  restrict   = restrict,
                                                  exact      = True)
            #print >>sys.stderr,"After LCheck"
     
            #
            # Build a bijective relationship for the left extensions
            #
            lk=list(lextensions.keys())
            for k in lk:
                #print >>sys.stderr,"Before LICheck",index.getRead(-lextensions[k][1][0],1,readLength-1)
                backe = index.checkedExtensions(probe      = index.getRead(-lextensions[k][1][0],1,readLength-1), 
                                                adapters5   = adapters5,
                                                adapters3   = adapters3,
                                                minread    = minread+delta, 
                                                minratio   = minratio, 
                                                mincov     = mincov,  
                                                minoverlap = minoverlap+delta, 
                                                extlength  = 1,
                                                lowfilter  = lowfilter,
                                                restrict   = restrict,
                                                exact      = True)
                #print >>sys.stderr,"After LICheck"
                
                # back = set(reduce(lambda p,q:p+q,[backe[y][1] for y in backe],[]))
                back = setofreads(backe)
                
                if readid not in back:
                    #print "Remove L extension : ",k
                    del lextensions[k]
                    
    #        print "Rext = ",rextensions
    #        print "Lext = ",lextensions
    
            if rextensions or lextensions:
                node  = graph.addNode(readid)
                node['gene']=gene
                if 'cycle' not in node:
                    node['cycle']=cycle
                if readid < readmax:
                    node['fake5']=0
                    node['fake3']=0
                else:
                    fake+=1
#                     node['graphics']={'fill':"#00FF00"}
    
    #        print "R",len(seq)
    
            for ex in rextensions:
                data = rextensions[ex]
                coverage = data[0]
    #            print ex,data
                
                nodeindexE,c,s = index.getIds(data[1][0])  # @UnusedVariable
                                    
                if nodeindexE in graph:
                    if data[1][0] not in s:
                        nodeindexE=-nodeindexE
                    # if readid!=-nodeindexE:
                    edges = graph.addEdge(readid,nodeindexE)
                    edges[1]['coverage']=coverage
                    edges[2]['coverage']=coverage
#                     edges[1]['label']="%s (%d)" % (edges[1]['ext'],coverage)
#                     edges[2]['label']="%s (%d)" % (edges[2]['ext'],coverage)
                    # else:
                    #     print >>sys.stderr,"\nWARNING : self loop on %d" % nodeindexE
        #              print "linked"
                else:
        #              print "not linked"
                    if not useonce or nodeindexE not in __used_reads__:
                        seeds.append((nodeindexE,(None,),int(delta)))
                        addedseed+=1
                                    
            #
            # extends the read on the left side
            # which is equivalent to extend on the right side
            # the reverse complement of the read
            #            
     
            for ex in lextensions:
                data = lextensions[ex]
                coverage = data[0]
    #            print ex,data
    
                nodeindexE,c,s = index.getIds(data[1][0])                               # @UnusedVariable
                                        
                if nodeindexE in graph:
                    if data[1][0] not in s:
                        nodeindexE=-nodeindexE
                    edges = graph.addEdge(-readid,nodeindexE)
                    edges[1]['coverage']=coverage
                    edges[2]['coverage']=coverage
#                     edges[1]['label']="%s (%d)" % (edges[1]['ext'],coverage)
#                     edges[2]['label']="%s (%d)" % (edges[2]['ext'],coverage)
                else:
                    if not useonce or nodeindexE not in __used_reads__:
                        seeds.append((nodeindexE,(None,),int(delta)))
                        addedseed+=1
            
            lastseeds.append((len(lextensions)+len(rextensions))/2.)
            if PyList_GET_SIZE(lastseeds) > 1000:
                lastseeds=lastseeds[-1000:]         

                
                        
    return graph



# cython: language_level=3

from cpython.function cimport PyFunction_Check

from ._graph cimport *

cdef str formatDict(dict keys, bint intfloat=False, 
                      str sep0="[", str sep1="]", 
                      str sep2="=", str sep3=" ", 
                      str sep4="[", sep5="]"):
    cdef str key
    cdef list output=[]
    
    for key,val in keys.iteritems():
        if isinstance(val, dict):
            output.append('%s%s%s%s%s%s%s' % (key,sep2,sep4, sep3,formatDict(val, intfloat, sep0, sep1, sep2, sep3, sep4, sep5),sep3,sep5))
        elif isinstance(val, bytes):
            output.append('%s%s"%s"' % (key,sep2,val.decode('ascii')))
        elif isinstance(val, bool):
            output.append('%s%s"%s"' % (key,sep2,val))
        elif intfloat and (isinstance(val, int) or isinstance(val, float)):
            output.append('%s%s%s' % (key,sep2,val))
        else:
            output.append('%s%s"%s"' % (key,sep2,val))
    
    return str("%s%s%s" % (sep0, sep3.join(output), sep1))


cdef class NodeIterator:

    def __init__(self,DiGraph graph, object predicate=None):
        
    #    if predicate is not None and not PyFunction_Check(predicate):
    #        raise TypeError,"predicate must be a function or None"
        
        self._nodepredicate = predicate
        self._graph = graph 
        self._exclude = set()
        self._nodeiterator = iter(graph._nodes)
        
    cpdef add(self,int node):
        cdef char[50] buffer
        cdef bytes key                          # @DuplicatedSignature
        
        snprintf(buffer,50,b"%d",node)
        key = buffer

        self._exclude.add(key)
        
    cpdef remove(self, int node):
        cdef char[50] buffer                    # @DuplicatedSignature
        cdef bytes key                          # @DuplicatedSignature
        
        snprintf(buffer,50,b"%d",node)
        key = buffer

        self.exclude.remove(key)
        
                
    def __next__(self):
        cdef bytes key
        cdef int node
        cdef bint ok
        
        
        ok =False
        
        while(not ok):
            key = next(self._nodeiterator)
            node = atol(key)
            ok = (self._nodepredicate is None or self._nodepredicate(node)) \
                   and key not in self._exclude
                
        return node
    
    def __ior__(self,others):
        cdef int i
        
        for i in others:
            i=i+0
            self.add(i)
            
        return self

    def __iter__(self):
        return self
    
    

cdef class DiGraph:

    def __init__(self,str name):
        self._nodes={}
        self._edges_attrs={}
        self._nodes_attrs={}
        self._name=name 
        
    def __contains__(self,int node):
        cdef char[50] buffer
        cdef bytes key
        
        snprintf(buffer,50,b"%d",node)
        key = buffer
        
        return key in self._nodes
    
    cpdef dict addNode(self,int node):
        cdef char[50] buffer    # @DuplicatedSignature
        cdef bytes key          # @DuplicatedSignature
        cdef dict node_attrs
        
        assert node < INT32_MAX,'Node id must be less than %d' % INT32_MAX
        assert node >= INT32_MIN,'Node id must be greater or equal than %d' % INT32_MIN
        
        snprintf(buffer,50,b"%d",node)
        key = buffer
        
        if key not in self._nodes:
            self._nodes[key]=(set(),set())
            node_attrs={}
            self._nodes_attrs[key]=node_attrs
        else:
            node_attrs=self._nodes_attrs[key]
            
        return node_attrs
            
    cpdef dict addEdge(self,int node1, int node2):
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edg
        cdef dict edges_attrs
        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
        
        edge = (key1,key2)
        
        if edge not in self._edges_attrs:
            self.addNode(node1)
            self.addNode(node2)
        
            self._nodes[key1][0].add(key2)
            self._nodes[key2][1].add(key1)

            edges_attrs={}
            self._edges_attrs[edge]=edges_attrs
        else:
            edges_attrs=self._edges_attrs[edge]
            
        return edges_attrs
            
    cpdef bint hasEdge(self, int node1, int node2):
        cdef char[50] buffer1   # @DuplicatedSignature
        cdef char[50] buffer2   # @DuplicatedSignature
        cdef bytes key1         # @DuplicatedSignature
        cdef bytes key2         # @DuplicatedSignature
        cdef tuple edgek
        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
   
        edgek = (key1,key2)

        return edgek in self._edges_attrs
    
    cpdef dict getEdgeAttr(self, int node1, int node2,int edge=-1):
 
        cdef char[50] buffer1           # @DuplicatedSignature
        cdef char[50] buffer2           # @DuplicatedSignature
        cdef bytes key1                 # @DuplicatedSignature
        cdef bytes key2                 # @DuplicatedSignature
        cdef tuple edgek                # @DuplicatedSignature
        cdef dict attr
        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
   
        edgek = (key1,key2)
        
        attr = self._edges_attrs[edgek]
        
        return attr 
    
    cpdef dict getNodeAttr(self, int node):
        cdef char[50] buffer        # @DuplicatedSignature
        cdef bytes key              # @DuplicatedSignature
        cdef dict attr              # @DuplicatedSignature
        
        snprintf(buffer,50,b"%d",node)
        key = buffer
        
        attr = self._nodes_attrs[key]
        
        return attr 

    cpdef int nodeCount(self):
        return len(self._nodes)
    
    cpdef int edgeCount(self):
        return len(self._edges_attrs)

    cpdef str gml(self,nodePredicate=None, edgePredicate=None):
        cdef str gml
        cdef int node1
        cdef int node2
        cdef int edge
        cdef list lines=[]
        cdef str line
        

        for node1 in self.nodeIterator(nodePredicate):
            lines.append('node [\n')
            line = "id %10d\n" % node1 
            lines.append(line)
            attr = self.getNodeAttr(node1)
            
            if attr:
                line = formatDict(attr, True, '', '', '\t\t', '\n')
                lines.append(line)

            lines.append(']')
            
        for node1,node2,edge in self.edgeIterator(nodePredicate,edgePredicate):
            lines.append('edge [\n')
            lines.append('source %d\n'%node1)
            lines.append('target %d\n'%node2)
 
 
            attr = self.getEdgeAttr(node1,node2,edge)
            assert isinstance(attr, dict)
            if attr:
                line = formatDict(attr, True, '', '', '\t\t', '\n')
            lines.append(line)
            lines.append(']')
        
        gml = "\n\t".join(lines)
        gml = 'Creator "%s"\nVersion "0.0"\ngraph [\n\tdirected\t1\n\tname\t"%s"\n%s\n]' % (__name__, self._name,gml)
        return gml

    
    cpdef str dot(self,nodePredicate=None, edgePredicate=None):
        cdef str dot
        cdef int node1              # @DuplicatedSignature
        cdef int node2              # @DuplicatedSignature
        cdef int edge               # @DuplicatedSignature
        cdef list lines=[]          # @DuplicatedSignature
        cdef str line             # @DuplicatedSignature
        cdef dict attr              # @DuplicatedSignature
        
        for node1 in self.nodeIterator(nodePredicate):
            attr = self.getNodeAttr(node1)
            if attr:
                line = "   %10d %s" % (node1,formatDict(attr))
            else:
                line = "   %10d" % node1 
            lines.append(line)
            
        for node1,node2,edge in self.edgeIterator(nodePredicate,edgePredicate):
            attr = self.getEdgeAttr(node1,node2,edge)
            if attr:
                line = "   %10d -> %10d %s" % (node1,node2,formatDict(attr))
            else:
                line = "   %10d -> %10d" % (node1,node2)
            lines.append(line)
        
        dot = "\n".join(lines)
        dot = "digraph %s {\n%s\n}" % (self._name,dot)
        
        return dot

    cpdef deleteEdge(self, int node1, int node2, int edge=-1):
        cdef char[50] buffer1   # @DuplicatedSignature
        cdef char[50] buffer2   # @DuplicatedSignature
        cdef bytes key1         # @DuplicatedSignature
        cdef bytes key2         # @DuplicatedSignature
        cdef tuple edgek        # @DuplicatedSignature

        
        snprintf(buffer1,50,b"%d",node1)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",node2)
        key2 = buffer2
   
        edgek = (key1,key2)
        
        del self._edges_attrs[edgek]
        
        self._nodes[key1][0].remove(key2)
        self._nodes[key2][1].remove(key1)
        
    cpdef deleteNode(self, int node):
        cdef char[50] buffer  # @DuplicatedSignature
        cdef bytes key  # @DuplicatedSignature
        cdef dict attr  # @DuplicatedSignature
        cdef int node2  # @DuplicatedSignature
        cdef list edges
        
        snprintf(buffer,50,b"%d",node)
        key = buffer
        
        edges = list(self.neighbourIterator(node))
        for node2 in edges:
            self.deleteEdge(node,node2)
            
        edges = list(self.parentIterator(node))
        for node2 in edges:
            self.deleteEdge(node2,node)
            
        del self._nodes[key]
        del self._nodes_attrs[key]
        
            
    cpdef DiGraph subgraph(self,str name, nodes):
        cdef DiGraph graph
        cdef set nodeset = set()
        cdef int node
        cdef int node2          # @DuplicatedSignature
        cdef char[50] buffer1   # @DuplicatedSignature
        cdef char[50] buffer2   # @DuplicatedSignature
        cdef bytes key1
        cdef bytes key2
        cdef tuple edge
        cdef set n0
        cdef set n1
        
        graph = self.__class__(name)
        
        for node in nodes:
            snprintf(buffer1,50,b"%d",node)
            key1 = buffer1
            nodeset.add(key1)
            
        for key1 in nodeset:
            n0 = self._nodes[key1][0]
            n1 = self._nodes[key1][1]
            
            graph._nodes[key1] = (n0 & nodeset, n1 & nodeset)
            graph._nodes_attrs[key1] = self._nodes_attrs[key1]
            
            for key2 in graph._nodes[key1][0]:
                edge=(key1,key2)                
                graph._edges_attrs[edge]=self._edges_attrs[edge]
        
        return graph
    
    cpdef set connectedComponent(self,int node, nodePredicate=None, edgePredicate=None):
        cdef set component=set([node])
        cdef set toexplore=set([node])
        cdef int n
        cdef int v 
        
        assert nodePredicate is None or nodePredicate(node)
        
        
        while toexplore:
            n = toexplore.pop()
            for v in self.neighbourIterator(n,nodePredicate, edgePredicate):
                if v not in component:
                    toexplore.add(v)
                    component.add(v)

            for v in self.parentIterator(n,nodePredicate, edgePredicate):
                if v not in component:
                    toexplore.add(v)
                    component.add(v)
                    
        return component
                
        
        
    
    def nodeIterator(self,predicate=None):
        return NodeIterator(self,predicate)
                
    def edgeIterator(self,nodePredicate=None, edgePredicate=None):
        cdef tuple edge
        cdef bytes key1
        cdef bytes key2
        cdef int   node1
        cdef int   node2
        
        for edge in self._edges_attrs:
            key1 = edge[0]
            key2 = edge[1]
            
            node1 = atol(key1)
            node2 = atol(key2)
            
            if nodePredicate is None or (nodePredicate(node1) and nodePredicate(node2)):
                edge = (node1,node2,0)
                if edgePredicate is None or edgePredicate(edge):
                    yield edge
                                
    def parentIterator(self,int node,nodePredicate=None, edgePredicate=None):
        cdef char[50] buffer
        cdef bytes key
        cdef set neighbours
        cdef int n 
        
        
        snprintf(buffer,50,b"%d",node)
        key = buffer
        
        neighbours = self._nodes[key][1]
        
        for key in neighbours:
            n = atol(key)
            if nodePredicate is None or nodePredicate(n):
                if edgePredicate is None or edgePredicate((n,node)):
                    yield n
                
    def neighbourIterator(self,int node,nodePredicate=None, edgePredicate=None):
        cdef char[50] buffer
        cdef bytes key
        cdef set neighbours
        cdef int n 
        
        
        snprintf(buffer,50,b"%d",node)
        key = buffer
        
        neighbours = self._nodes[key][0]
        
        for key in neighbours:
            n = atol(key)
            if nodePredicate is None or nodePredicate(n):
                if edgePredicate is None or edgePredicate((node,n)):
                    yield n
                    

    def connectedComponentIterator(self,nodePredicate=None, edgePredicate=None):
        cdef NodeIterator ni=self.nodeIterator(nodePredicate)
        cdef set cc
        cdef int n
        
        for n in ni:
            cc = self.connectedComponent(n,nodePredicate,edgePredicate)
            ni|=cc
            yield cc
            
        
        
    def __len__(self):
        return self.nodeCount()
    
    def __iter__(self):
        return self.nodeIterator()
            
    
    def __getitem__(self,key):
        cdef int   node1
        cdef int   node2
                
        if isinstance(key,tuple):
            node1 = key[0]
            node2 = key[1]
            
            return self.getEdgeAttr(node1,node2)
        
        elif isinstance(key,int):
            node1 = key

            try : 
                return self.getNodeAttr(node1)
            except IndexError as e:
                raise IndexError("%s --> index requested %d : %d" % (e.args[0],node1,key))
            
        raise TypeError("key must be an int or a tuple")

    def __str__(self):
        return self.dot()
    
    def __delitem__(self, int node):
        self.deleteNode(node)
        
    def minpath(self,source,dest = None, distance=None, nodePredicate=None, edgePredicate=None, allowCycle=False):
        cdef dict dist
        cdef dict previous
        cdef int32_t v 
        cdef FibonacciMinIntHeap Q = FibonacciMinIntHeap(chunksize=self.nodeCount()+1)
        cdef size_t d 
        cdef int32_t u
        cdef set idest
        
        def defaultdist(int32_t a, int32_t b):
            return 1
        
        if distance is None:
            distance=defaultdist
            
        dist={}
        previous={}
        
        for v in self.nodeIterator(nodePredicate):
            Q.push(v,INT32_MAX)
           
        idest = None 
        if dest is not None:
            idest=set(dest)
            
        if isinstance(source, int):
            source=set([source])
        else:
            source=set(source)
            
        for s in source:
            Q[s]=0
        
        while Q and (idest is None or dest):
            u,d = Q.pop()
            
            if d == INT32_MAX or (idest is not None and not idest):
                break
            
            dist[u]=d
            
            for v in self.neighbourIterator(u,nodePredicate, edgePredicate):
                if v in Q:
                    alt = d + distance(u,v)
                    if alt < Q.getPriority(v):
                        Q[v]=alt
                        previous[v]=u
                elif allowCycle and v in source:
                    alt = d + distance(u,v)
                    if alt < dist[v] or dist[v]==0:
                        dist[v]=alt
                        previous[v]=u
            
            if idest is not None and u in idest:
                idest.remove(u)

        idest = None 
        if dest is not None:
            idest=set(dist) - set(dest)
            for i in idest:
                del dist[i]
            
                        
        return dist,previous
                    
                    
    def __bytes__(self):
        return bytes(self.__str__.decode('latin1'))
        

def pathIterator(dist,previous):
    
    for p in dist:
        path=[p]
        while (p in previous):
            p= previous[p]
            path.append(p)
        path.reverse()
        yield path
        
                
        

        


'''
Created on 20 oct. 2012

@author: coissac
'''

import os

from obidistutils.command.build_exe import build_exe
from obidistutils.serenity.checksystem import is_mac_system


class pidname(build_exe):
    
    description = "build the pidname executable returning the executable path from a PID on a mac"


    def initialize_options(self):
        build_exe.initialize_options(self)

        self.pidname = False


    def finalize_options(self):
        # This might be confusing: both build-cexe and build-temp default
        # to build-temp as defined by the "build" command.  This is because
        # I think that C libraries are really just temporary build
        # by-products, at least from the point of view of building Python
        # extensions -- but I want to keep my options open.
        
        build_exe.finalize_options(self)
        
        self.set_undefined_options('build',
                                   ('build_scripts', 'build_cexe'))

        # self.ctools = self.distribution.ctools
        
        if os.path.isfile("distutils.ext/src/pidname.c"):
            self.executables = [('pidname',{"sources":["distutils.ext/src/pidname.c"]})]
            self.check_executable_list(self.executables)
        else:
            self.executables = []

#         self.build_cexe = os.path.join(os.path.dirname(self.build_cexe),'cbinaries') 
#         self.mkpath(self.build_cexe)



    def run(self):
        if is_mac_system():
            build_exe.run(self)
            self.pidname=True
        else:
            self.pidname=False
        


# cython: language_level=3

from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_GET_SIZE
from collections import Counter

#                   "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
cdef char* __COMP = "T-G---C------------A------"

cdef int _reverseComplement(char* seq, size_t length):
    """
    Internal function computing the reverse complement sequence
    of a DNA sequence in place.
    """
    cdef char   tmp
    cdef char*  end = seq + length -1
    
    while seq <= end:
        tmp = end[0]
        end[0] = __COMP[(seq[0] & (~32)) - 65]
        seq[0] = __COMP[(tmp & (~32)) - 65]
        seq+=1
        end-=1 
        
    return 0
    
cpdef bytes reverseComplement(bytes seq):
    """
    Computes the reverse complement sequence of a DNA sequence
    composed only of acgt. No check is done for efficiency 
    purpose so be careful to the input sequence.
    
    :param seq: the sequence to revert-compelent expressed as
                a serie of acgt in upper or lower case
    :type  seq: a bytes instance
    
    :return: the reverse-complemented sequence in upper case
    :rtype:  a bytes instance
    """
    cdef bytes  comp 
    cdef char*  ccomp= seq
    cdef size_t lcomp= PyBytes_GET_SIZE(seq)
    
    comp = PyBytes_FromStringAndSize(ccomp,lcomp)
    ccomp= comp
    
    _reverseComplement(ccomp,lcomp)
    
    return comp

cpdef bint isDNA(bytes seq):
    cdef set c = set([x for x in seq.upper()])
    return len(c)<=4


cdef dict codon = {}
cdef int i

ncbieaa="FFLLSSSSYYQQCCGWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG"
Base1  ="TTTTTTTTTTTTTTTTCCCCCCCCCCCCCCCCAAAAAAAAAAAAAAAAGGGGGGGGGGGGGGGG"
Base2  ="TTTTCCCCAAAAGGGGTTTTCCCCAAAAGGGGTTTTCCCCAAAAGGGGTTTTCCCCAAAAGGGG"
Base3  ="TCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAG"

for i in range(len(ncbieaa)):
    codon[Base1[i]+Base2[i]+Base3[i]]=ncbieaa[i]
    
cpdef tuple translate3(bytes seq): 
    cdef int i                               # @DuplicatedSignature
    cdef list prot0=[]
    cdef list prot1=[]
    cdef list prot2=[]
    
    
    for i in range(len(seq)-3):
        aa = codon[seq[i:(i+3)].upper()]
        p = i % 3
        if p==0:
            prot0.append(aa)
        elif p==1:
            prot1.append(aa)
        else:
            prot2.append(aa)
            
    return (b''.join(prot0),
            b''.join(prot1),
            b''.join(prot2)
            )
            


'''
Created on 2 oct. 2014

@author: coissac
'''

import re
import os

try:
    import pip                                              # @UnresolvedImport
    from pip.utils import get_installed_distributions       # @UnresolvedImport
except ImportError:
    from .bootstrappip import bootstrap
    bootstrap()

from distutils.version import StrictVersion             # @UnusedImport
from distutils.errors import DistutilsError
from distutils import log

class RequirementError(Exception):
    pass

def is_installed(requirement):
    
    
    requirement_project,requirement_relation,requirement_version = parse_package_requirement(requirement)
    
    package = [x for x in get_installed_distributions() if x.project_name==requirement_project]
    
    if len(package)==1:
        if requirement_version is not None and requirement_relation is not None:    
            rep = (len(package)==1) and eval("StrictVersion('%s') %s StrictVersion('%s')" % (package[0].version,
                                                                                           requirement_relation,
                                                                                           requirement_version)
                                             )
        else:
            rep=True
    else:
        rep=False
    
    if rep:
        if requirement_version is not None and requirement_relation is not None:        
            log.info("Look for package %s (%s%s) : ok version %s installed" % (requirement_project,
                                                                               requirement_relation,
                                                                               requirement_version,
                                                                               package[0].version))
        else:
            log.info("Look for package %s : ok version %s installed" % (requirement_project,
                                                                        package[0].version))
    else:
        if len(package)!=1:
            log.info("Look for package %s (%s%s) : not installed" % (requirement_project,
                                                                     requirement_relation,
                                                                     requirement_version))
        else:
            log.info("Look for package %s (%s%s) : failed only version %s installed" % (requirement_project,
                                                                                        requirement_relation,
                                                                                        requirement_version,
                                                                                        package[0].version))
        
    return rep


def get_requirements(requirementfile='requirements.txt'):
    
    try:
        requirements = open(requirementfile).readlines()
        requirements = [x.strip() for x in requirements]
        requirements = [x for x in requirements if x[0]!='-']
    
    except IOError:
        requirements = []
        
    return requirements
    
    
def install_requirements(requirementfile='requirements.txt'):
    
    install_something=False

    requirements = get_requirements(requirementfile)

    log.info("Required packages for the installation :")
    for x in requirements:
        ok = is_installed(x)
        if not ok:
            log.info("  Installing requirement : %s" % x)
            pip_install_package(x)
            install_something=True
            if x[0:3]=='pip':
                return True
                
    return install_something
 

def check_requirements(requirementfile='requirements.txt'):
    
    requirements = get_requirements(requirementfile)
    
    log.info("Required packages for the installation :")
    for x in requirements:
        ok = is_installed(x)
        if not ok:
            raise RequirementError("  Missing requirement : %s -- Package installation stopped" % x)
                
 


def parse_package_requirement(requirement):
    
    version_pattern = re.compile('[=><]+(.*)$')
    project_pattern  = re.compile('[^=><]+')
    relationship_pattern = re.compile('[=><]+')
    
    try:
        requirement_project = project_pattern.search(requirement).group(0)
        requirement_version = version_pattern.search(requirement)
        if requirement_version is not None:
            requirement_version=requirement_version.group(1)
        requirement_relation= relationship_pattern.search(requirement)
        if requirement_relation is not None:
            requirement_relation=requirement_relation.group(0)
    except:
        raise DistutilsError("Requirement : %s not correctly formated" % requirement)
    
    return requirement_project,requirement_relation,requirement_version
    

def get_package_requirement(package,requirementfile='requirements.txt'):            
    requirements = get_requirements(requirementfile)
    req = [x for x in requirements
             if x[0:len(package)]==package
          ]
    
    if len(req)==1:
        return req[0]
    else:
        return None
        
        
def pip_install_package(package,directory=None,upgrade=True):

    if directory is not None:
        log.info('    installing %s in directory %s' % (package,str(directory)))
        

    if 'http_proxy' in os.environ and 'https_proxy' not in os.environ:
        os.environ['https_proxy']=os.environ['http_proxy']

    args = ['install']
    
    if upgrade:
        args.append('--upgrade')
    
    if 'https_proxy' in os.environ:
        args.append('--proxy=%s' % os.environ['https_proxy'])
        
    if directory is not None:
        args.append('--target=%s' % directory)
    
    args.append(package)
        
    return pip.main(args)




'''
Created on 20 oct. 2012

@author: coissac
'''

import os.path

from distutils.command.build_scripts import build_scripts as ori_build_scripts
from distutils.util import convert_path
from distutils import log, sysconfig
from distutils.dep_util import newer
from stat import ST_MODE
import re

first_line_re = re.compile('^#!.*python[0-9.]*([ \t].*)?$')

class build_scripts(ori_build_scripts):
            
    def copy_scripts (self):
        """Copy each script listed in 'self.scripts'; if it's marked as a
        Python script in the Unix way (first line matches 'first_line_re',
        ie. starts with "\#!" and contains "python"), then adjust the first
        line to refer to the current Python interpreter as we copy.
        """        
        self.mkpath(self.build_dir)
        rawbuild_dir = os.path.join(os.path.dirname(self.build_dir),'raw_scripts')
        self.mkpath(rawbuild_dir)
        
        outfiles = []
        for script in self.scripts:
            adjust = 0
            script = convert_path(script)
            outfile = os.path.join(self.build_dir, os.path.splitext(os.path.basename(script))[0])
            rawoutfile = os.path.join(rawbuild_dir, os.path.basename(script))
            outfiles.append(outfile)

            if not self.force and not newer(script, outfile):
                log.debug("not copying %s (up-to-date)", script)
                continue

            # Always open the file but ignore failures in dry-run mode --
            # that way, we'll get accurate feedback if we can read the
            # script.
            try:
                f = open(script, "r")
            except IOError:
                if not self.dry_run:
                    raise
                f = None
            else:
                first_line = f.readline()
                if not first_line:
                    self.warn("%s is an empty file (skipping)" % script)
                    continue

                match = first_line_re.match(first_line)
                if match:
                    adjust = 1
                    post_interp = match.group(1) or ''
            
            log.info("Store the raw script %s -> %s", script,rawoutfile)        
            self.copy_file(script, rawoutfile)


            if adjust:
                log.info("copying and adjusting %s -> %s", script,
                         self.build_dir)
                if not self.dry_run:
                    outf = open(outfile, "w")
                    if not sysconfig.python_build:
                        outf.write("#!%s%s\n" %
                                   (self.executable,
                                    post_interp))
                    else:
                        outf.write("#!%s%s\n" %
                                   (os.path.join(
                            sysconfig.get_config_var("BINDIR"),
                           "python%s%s" % (sysconfig.get_config_var("VERSION"),
                                           sysconfig.get_config_var("EXE"))),
                                    post_interp))
                    outf.writelines(f.readlines())
                    outf.close()
                if f:
                    f.close()
            else:
                if f:
                    f.close()
                self.copy_file(script, outfile)

        if os.name == 'posix':
            for F in outfiles:
                if self.dry_run:
                    log.info("changing mode of %s", F)
                else:
                    oldmode = os.stat(F)[ST_MODE]
                    oldmode = oldmode & 0o7777
                    
                    newmode = (oldmode | 0o555) & 0o7777
                    if newmode != oldmode:
                        log.info("changing mode of %s from %o to %o",
                                 F, oldmode, newmode)
                        os.chmod(F, newmode)




# cython: language_level=3

cdef extern from "fibo.h":
    ctypedef struct FiboNode:
        pass
    
    ctypedef struct FiboTree:
        pass
    
    ctypedef FiboNode* const_FiboNode_ptr_const "const FiboNode * const"
    ctypedef FiboNode* const_FiboNode_ptr "const FiboNode *"
    ctypedef FiboTree* const_FiboTree_ptr "const FiboTree *"
    
    int fiboTreeInit(const_FiboTree_ptr tree, int (*) (const_FiboNode_ptr_const node1,const_FiboNode_ptr_const node2))
    void fiboTreeExit(const_FiboTree_ptr tree)
    void fiboTreeFree(const_FiboTree_ptr tree)
    FiboNode *fiboTreeConsolidate (const_FiboTree_ptr tree)
    void fiboTreeAdd(const_FiboTree_ptr tree, const_FiboNode_ptr node)
    void fiboTreeDel(const_FiboTree_ptr tree, const_FiboNode_ptr node)
    FiboNode *fiboTreeMin(const_FiboTree_ptr tree)
    int fiboTreeCheck(const_FiboTree_ptr tree)



'''
Created on 27 sept. 2016

@author: coissac
'''
from orgasm import getOutput, getIndex, getSeeds
from orgasm.tango import restoreGraph, estimateFragmentLength, genesincontig,\
    scaffold, pairEndedConnected
    
from orgasm.version import version
import pathlib
import sys

__title__="Build a graph file from the assembling graph"

default_config = { 'format': 'gml'
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='<index>', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='<output>', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    parser.add_argument('--gml',              dest='graph:format',
                                              action='store_const',
                                              const='gml',
                                              default=None,
                        help="Write the assembling graph in gml format [default]")
        
    
    parser.add_argument('--gml-path',         dest='graph:format',
                                              action='store_const',
                                              const='gmlpath',
                                              default=None,
                        help="Write the assembling graph in gml format with "
                             "the indication of the last exported sequence path")
        
    parser.add_argument('--lastgraph',        dest='graph:format',
                                              action='store_const',
                                              const='lastgraph',
                                              default=None,
                        help="Write the assembling graph in lastgraph format (Velvet). "
                             "This format is readable by the Bandage visualizer")
        
    
    parser.add_argument('--back',             dest='orgasm:back', 
                                              metavar='<insert size>',
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with pared-ends [default: <estimated>]')



def compactGraph2lastgraph(asm,paths=[],minlink=5,back=250,logger=None):
#FASTG:begin;
#FASTG:version=1.0:assembly_name=”tiny example”;
    cg = asm.compactAssembling(verbose=False)
    genesincontig(cg,asm.index,asm.seeds)

    scaffold(asm,
             cg,
             minlink=minlink,
             back=int(back),
             addConnectedLink=False,
             logger=logger)
    
    edges = list(cg.stemIterator())
    nodes = [e for e in edges if "stemid" in e]
    seqcount = sum(e['length'] for e in nodes if 'gappairs' not in e) / 2
    
    nodes = dict((n['stemid'],n) for n in nodes)
    
    # Print the header of the file
    
    print("{nodecount:d} {seqcount:d} {hash:d} 1".format(
                nodecount = int(len(nodes)/2),
                seqcount  = int(seqcount),
                hash      = int(asm.index.getReadSize())
            )
          )
 
    begins  = {}
 
    # print the nodes -- here nodes corresponds to edge in oa
    for n in range(1,int(len(nodes)/2)+1): 
        first = nodes[n]['first']
        if first not in begins:
            begins[first]=[]
        begins[first].append(n)

        first = nodes[-n]['first']
        if first not in begins:
            begins[first]=[]
        begins[first].append(-n)
        
               
        print("NODE    {NODE_ID:d}  {LENGTH:d}  {COV_SHORT1:d}   {O_COV_SHORT1:d}  {COV_SHORT2:d}  {O_COV_SHORT2:d}".format(
                NODE_ID=n,
                LENGTH=len(nodes[n]['sequence']),
                COV_SHORT1=nodes[n]['weight'] * len(nodes[n]['sequence']),
                O_COV_SHORT1=nodes[n]['weight'] * len(nodes[n]['sequence']),
                COV_SHORT2=0,
                O_COV_SHORT2=0
             )
            )
        print(nodes[n]['sequence'].decode('ASCII'))
        print(nodes[-n]['sequence'].decode('ASCII'))
        
    # Print arc joining nodes
    done=set()
    for n in range(1,int(len(nodes)/2)+1): 
        last = nodes[n]['last']
        if last in begins:
            for ne in begins[last]:
                if 'path' not in nodes[n]:
                    connected=nodes[n]['pairendlink']
                elif 'path' not in nodes[ne]:
                    connected=nodes[ne]['pairendlink']
                else:
                    connected,ml,sl,delta = pairEndedConnected(asm,cg,n,ne,back)  # @UnusedVariable
                done.add((n,ne))
                print("ARC {START_NODE}    {END_NODE}    {MULTIPLICITY}".format(
                            START_NODE=n,
                            END_NODE=ne,
                            MULTIPLICITY=connected
                        )
                      )
        last = nodes[-n]['last']
        if last in begins:
            for ne in begins[last]:
                if (-ne,-n) not in done:
                    if 'path' not in nodes[n]:
                        connected=nodes[-n]['pairendlink']
                    elif 'path' not in nodes[ne]:
                        connected=nodes[ne]['pairendlink']
                    else:
                        connected,ml,sl,delta = pairEndedConnected(asm,cg,-n,ne,back)  # @UnusedVariable
                    print("ARC {START_NODE}    {END_NODE}    {MULTIPLICITY}".format(
                                START_NODE=-n,
                                END_NODE=ne,
                                MULTIPLICITY=connected
                            )
                          )
            

def compactGraph2gfa(cg,paths=[]):

    print("# Organelle Assembler - version {}".format(version))
    print("# Graphical Fragment Assembly (GFA) file")
    print('H    VN:Z:1.0')
    
    begins = {}
    ends = {}
    for node1,node2,edge in cg.edgeIterator():
        stem = cg.getEdgeAttr(node1,node2,edge)
        
        if 'stemid' in stem:
            bl = begins.get(node1,[])
            begins[node1]=bl
            bl.append(stem["stemid"])
    
            ends[stem["stemid"]]=node2
    
            if (stem["stemid"] > 0) :    
                print("S    {:6d}    {}    RC:i:{}".format(
                            stem["stemid"],
                            (stem['head'] + stem['sequence']).decode("ASCII"),
                            stem["weight"]
                        )
                      )
            
    for stemid1 in ends:
        if stemid1 > 0:
            if ends[stemid1] in begins:
                for stemid2 in begins[ends[stemid1]]:
                    if stemid2 > 0:
                        print('L    {:6d} + {:6d} + *'.format(
                                    stemid1,
                                    stemid2
                                )
                              )

    print(begins)
    print(ends)
    
def cat(filename):
    with open(filename,'r') as f:
        for line in f:
            print(line)
            
def run(config):

    logger=config['orgasm']['logger']
    output = getOutput(config)


    if not pathlib.Path(output+'.gml').is_file():
        logger.error("No assembly graph available")
        sys.exit(1)
        
    if config['graph']['format']=='gml':
        cat(output+'.gml')
    elif  config['graph']['format']=='gmlpath':
        if pathlib.Path(output+'.path.gml').is_file():
            cat(output+'.path.gml')
        else:
            cat(output+'.gml')
    elif  config['graph']['format']=='lastgraph':
        r = getIndex(config)
        ecoverage,seeds,newprobes = getSeeds(r,config)  # @UnusedVariable
        
        asm = restoreGraph(output+'.oax',r,seeds)
    
        logger.info("Evaluate fragment length")
        
        meanlength,sdlength = estimateFragmentLength(asm)
        
        if meanlength is not None:
            logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))
    
        if config['orgasm']['back'] is not None:
            back = config['orgasm']['back']
        elif config['orgasm']['back'] is None and meanlength is not None:
            back = int(meanlength + 4 * sdlength)
            if back > 500:
                back=500
        else:
            back = 300
        compactGraph2lastgraph(asm,logger=logger)




'''
Created on 28 sept. 2014

@author: coissac
'''
from orgasm import getIndex

__title__="List information about a read index"

default_config = { 
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    


def run(config):

    r = getIndex(config)
    
    print(len(r),r.getReadSize())

              



'''
Created on 4 sept. 2014

@author: coissac
'''

from ConfigParser import SafeConfigParser, NoOptionError
import logging
from os import access
from os import R_OK
import sys
import argparse


def getLogger(config):
    
    rootlogger   = logging.getLogger()
    logFormatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")

    stderrHandler = logging.StreamHandler(sys.stderr)
    stderrHandler.setFormatter(logFormatter)

    rootlogger.addHandler(stderrHandler)
    
    if config.log:
        fileHandler = logging.FileHandler("%s.log" % config.outputname)
        fileHandler.setFormatter(logFormatter)
        rootlogger.addHandler(fileHandler)
    
    try:
        loglevel = getattr(logging, config.get('General','loglevel')) 
    except:
        loglevel = logging.INFO
        
    rootlogger.setLevel(loglevel)
    
    return rootlogger

def str2loglevel(level):
    try:
        level = getattr(logging,level)
    except AttributeError:
        level = logging.INFO
        
    return level  

class Config:
    
    def __init__(self,config):
        self._config=config
        self._key = {}
        self._isValued = {}
        
        self._logger=getLogger(self)
        
    def _getvalue(self,key,section,option,vtype=str):
        if not self._isValued.get('key',False):
            try:
                self._key[key]=vtype(self._config.get(section,option))
            except NoOptionError:
                self._key[key]=None
            self._isValued[key]= True
        return self._key[key]
        
    @property
    def logger(self):
        """
        The logger object usable by the assembler
        """
        return self._logger
    
    @property
    def log(self):
        """
        The file name of the log file. If None then logs are not saved in a file
        """
        log = self._getvalue('log','General','log',bool)
        return log
        
    @property
    def loglevel(self):
        """
        The file name of the log file. If None then logs are not saved in a file
        """
        loglevel = self._getvalue('loglevel','General','loglevel',str2loglevel)
        return loglevel
   

    @property
    def output(self):
        output = self._getvalue('output','General','#@#@#@',str)
        if output is None:
            if self.outputname is not None:
                output = open(self.outputname(),'w')
            else:
                output = sys.stdout
            self._isValued['output']= True
            self._key['output']=output
        return output

    @property
    def seedname(self):
        seedname = self._getvalue('seedname','Probes','seeds',str)
        return seedname
        
    @property
    def indexname(self):
        indexname = self._getvalue('indexname','Data','index',str)
        return indexname
    
    @property
    def outputname(self):
        outputname = self._getvalue('outputname','Data','output',str)
        return outputname
    
    @property
    def seedmincov(self):
        mincov = self._getvalue('mincov','Probes','mincov',int)
        return mincov if mincov > 0 else 1
    
    @property
    def minread(self):
        """
        The minimum count of usable reads to consider an extension
        """
        
        minread = self._getvalue('minread','Extension','minread',int)
        return minread if minread > 0 else None
    
    @property
    def coverage(self):
        """
        The expected sequencing coverage
        """
        
        coverage = self._getvalue('coverage','Extension','coverage',int)
        return coverage if coverage > 0 else None
    
    @property
    def mincov(self):
        """
        The minimum count of observation of a read to consider it during an extension
        """
        mincov = self._getvalue('mincov','Extension','mincov',int)
        return mincov if mincov > 0 else 1
    
    @property
    def minoverlap(self):
        minoverlap = self._getvalue('minoverlap','Extension','minoverlap',int)
        return minoverlap if minoverlap > 0 else 1
    
    @property
    def minratio(self):
        minratio = self._getvalue('minratio','Extension','minratio',float)
        return minratio
    
    @property
    def lowcomplexity(self):
        lowcomplexity = self._getvalue('minratio','Extension','minratio',bool)
        return lowcomplexity
        
    @property
    def back(self):        
        back = self._getvalue('back','Gap-filling','back',int)
        return back if back > 0 else None
    
    @property
    def snp(self):
        snp = self._getvalue('snp','Cleaning','snp',bool)
        return snp
    
    @property
    def smallbranches(self):
        """
        The minimum count of usable reads to consider an extension
        """
        smallbranches = self._getvalue('smallbranches','Cleaning','smallbranches',int)
        return smallbranches if smallbranches > 0 else None
    

    

def getConfigFile(options=None):
    
    configFile = None
    if options is not None:
        if hasattr(options, 'config'):
            if access(options.config,R_OK):
                configFile = options.config
            else:
                raise IOError("Config file : %s not readable" % options.config)
    
    if configFile is not None:
        if access("orgasm.conf",R_OK):
            configFile = "orgasm.conf"
        elif access("~/.orgasm.conf",R_OK):
            configFile = "~/.orgasm.conf"
            
    return configFile

def defaultConfiguration(options=None):
    
    config = SafeConfigParser()
    config.optionxform=str

    config.add_section('General')
    config.set('General', 'log', 'False')
    config.set('General', 'loglevel', 'INFO')
    
    config.add_section('Data')
    
    config.add_section('Probes')
    config.set('Probes', 'seeds',  'protChloroArabidopsis')
    config.set('Probes', 'mincov', '1')
   
    config.add_section('Extension')
    config.set('Extension', 'minread',       '-1')
    config.set('Extension', 'coverage',      '-1')
    config.set('Extension', 'mincov',        '1')
    config.set('Extension', 'minratio',      '0.1')
    config.set('Extension', 'minoverlap',    '50')
    config.set('Extension', 'lowcomplexity', 'False')

    config.add_section('Cleaning')
    config.set('Cleaning', 'smallbranches', '-1')
    config.set('Cleaning', 'snp',           'True')

    config.add_section('Gap-filling')
    config.set('Gap-filling', 'back', '-1')
   
    
    return config

def getConfiguration(options=None):
    
    
    dconfig = defaultConfiguration()
    configFile = getConfigFile(options)
    
    if configFile is not None:
        dconfig.read(configFile)
        
    if options is not None:
        if hasattr(options,'minread') and options.minread is not None:
            dconfig.set('Extension', 'minread', str(options.minread))
    
        if hasattr(options,'coverage') and options.coverage is not None:
            dconfig.set('Extension', 'coverage', str(options.coverage))
    
        if hasattr(options,'minratio') and options.minratio is not None:
            dconfig.set('Extension', 'minratio', str(options.minratio))
    
        if hasattr(options,'mincov') and options.mincov is not None:
            dconfig.set('Extension', 'mincov', str(options.mincov))

        if hasattr(options,'minoverlap') and options.minoverlap is not None:
            dconfig.set('Extension', 'minoverlap', str(options.minoverlap))

        if hasattr(options,'smallbranches') and options.smallbranches is not None:
            dconfig.set('Cleaning', 'smallbranches', str(options.smallbranches))

        if hasattr(options,'outputFilename') and options.outputFilename is None:
            options.outputFilename = options.indexFilename
        
        dconfig.set('Data', 'index', options.indexFilename)
        dconfig.set('Data', 'output', options.outputFilename)
            
        if options.log:
            dconfig.set('General', 'log', True)
        
        if options.back is not None:
            dconfig.set('Gap-filling', 'back', str(options.back))
    
        if options.seeds is not None:
            dconfig.set('Probes', 'seeds', str(options.seeds))
   
        
        
    return Config(dconfig)



#cython: language_level=3

'''
Created on 27 mars 2016

@author: coissac
'''

import sys

from .command   import  getCommandsList
from .logging   cimport getLogger
from .arguments cimport buildArgumentParser

from ..version import version
from _curses import version

cdef dict __default_config__ = {}


cpdef str setRootConfigName(str rootname):
    global __default_config__
    if '__root_config__' in __default_config__:
        if __default_config__["__root_config__"] in __default_config__:
            __default_config__[rootname]=__default_config__[__default_config__["__root_config__"]]
            del __default_config__[__default_config__["__root_config__"]]
    __default_config__['__root_config__']=rootname
    return rootname

cpdef str getRootConfigName():
    global __default_config__
    return __default_config__.get('__root_config__',None)

cdef dict buildDefaultConfiguration(str root_config_name,
                                    dict  config):
    global __default_config__
    
    __default_config__.clear()
    setRootConfigName(root_config_name)    

    __default_config__[root_config_name]=config
    
    config['version']=version
    
    commands = getCommandsList()
    
    for c in commands:
        module = commands[c]
    
        assert hasattr(module, "run")
        
        if hasattr(module, 'default_config'):
            __default_config__[c]=module.default_config
        else:
            __default_config__[c]={}
                        
    return __default_config__
        

cpdef dict getConfiguration(str root_config_name="__default__",
                            dict  config={}):
    global __default_config__

    if '__done__' in __default_config__:
        return __default_config__
    
    if root_config_name=="__default__":
        raise RuntimeError("No root_config_name specified")
    
    if not config:
        raise RuntimeError("Base configuration is empty")
    
    
    
    config =  buildDefaultConfiguration(root_config_name,
                                        config)
    
    parser = buildArgumentParser(root_config_name,
                                 config[root_config_name]['software'])
    
    options = vars(parser.parse_args())

    if options['%s:version' % root_config_name]:
        print("%s - Version %s" % (config[root_config_name]['software'],
                                   config[root_config_name]['version']))
        sys.exit(0)
    
    for k in options:
        section,key = k.split(':')
        s = config[section]
        if options[k] is not None:
            s[key]=options[k]
            
    if not 'module' in config[root_config_name]:
        print('\nError: No command specified',file=sys.stderr)
        parser.print_help()
        sys.exit(2)
                
    getLogger(config)
    
    config['__done__']=True
            
    return config



'''
Created on 10 mars 2015

@author: coissac
'''
from distutils.core import Command
import os.path
import glob

class install_sphinx(Command):
    '''
    Install the sphinx documentation
    '''
    
    description = "Install the sphinx documentation in serenity mode"

    boolean_options = ['force', 'skip-build']


    def initialize_options (self):
        self.install_doc = None
        self.build_dir = None

    def finalize_options (self):
        self.set_undefined_options('build_sphinx', ('build_dir', 'build_dir'))
        self.set_undefined_options('install',
                                   ('install_scripts', 'install_doc'))

    def run (self):
        if self.distribution.serenity:
            self.install_doc = os.path.join(self.install_doc,"../export/share")
            self.install_doc=os.path.abspath(self.install_doc)
            self.mkpath(self.install_doc)
            self.mkpath(os.path.join(self.install_doc,'html'))
            outfiles = self.copy_tree(os.path.join(self.build_dir,'html'),  # @UnusedVariable
                                      os.path.join(self.install_doc,'html'))
                
            self.mkpath(os.path.join(self.install_doc,'man','man1'))
            outfiles = self.copy_tree(os.path.join(self.build_dir,'man'),  # @UnusedVariable
                                      os.path.join(self.install_doc,'man','man1'))

            for epub in glob.glob(os.path.join(self.build_dir,'epub/*.epub')):
                self.copy_file(os.path.join(epub), 
                               os.path.join(self.install_doc,os.path.split(epub)[1]))
            
    def get_outputs(self):
        directory=os.path.join(self.install_doc,'html')
        files = [os.path.join(self.install_doc,'html', f) 
                 for dp, dn, filenames in os.walk(directory) for f in filenames]  # @UnusedVariable
        
        directory=os.path.join(self.build_dir,'man')
        files.append(os.path.join(self.install_doc,'man','man1', f) 
                 for dp, dn, filenames in os.walk(directory) for f in filenames)  # @UnusedVariable

        directory=os.path.join(self.build_dir,'epub')
        files.append(os.path.join(self.install_doc, f) 
                 for dp, dn, filenames in os.walk(directory)  # @UnusedVariable
                 for f in glob.glob(os.path.join(dp, '*.epub')) )
        
        return files
        


#cython: language_level=3

'''
Created on 25 mars 2016

@author: coissac
'''

from urllib.request import urlopen


    
def uopen(str name, mode='r', int buffersize=100000000):
    cdef CompressedFile c
    cdef LineBuffer lb
    
    try:
        f = urlopen(name)
    except:
        f = open(name,mode) 
        
    c = CompressedFile(f)
    
    if isinstance(c, LineBuffer):
        lb=c
    else:
        lb=LineBuffer(c,buffersize)
        
    i = iter(lb)
    
    return i



# cython: language_level=3

cimport cython

from orgasm.indexer._orgasm cimport *
from ._asmbgraph cimport *
from ._assembler cimport *
from orgasm.graph._graphmultiedge cimport *
from functools import reduce

from orgasm.apps.progress cimport ProgressBar 
from orgasm.apps.config cimport getConfiguration 
from orgasm.utils.dna cimport reverseComplement
import math

import sys

from cpython.sequence cimport PySequence_GetSlice
from cpython.bytes cimport PyBytes_GET_SIZE


def cmp(a, b):
    return (a > b) - (a < b)

cdef int iabs(int x):
    return x if x > 0 else -x

@cython.nonecheck(True)
cdef tuple findDeepestRead(Index index, bytes seed):
    cdef list frg
    cdef list fcount
    cdef int mfc
    cdef int readsize=index.getReadSize()
    cdef int x
    cdef bytes b
    
    frg   = [PySequence_GetSlice(seed,x,(x+readsize)) for x in range(PyBytes_GET_SIZE(seed)-readsize)]
    fcount= [index.count(b) for b in frg]
    mfc = max(fcount)
    return frg[fcount.index(mfc)],mfc

cpdef cmpPoints(p1,p2):
    '''
    Compare two extensions point.
    INTERNAL functio uses for ordering extension point:
    the smallest point is point closest to the initial extension point
    with the highest coverage.
    
    :param p1:
    :type p1:
    :param p2:
    :type p2:
    '''
    
    rep = cmp(p2[0],p1[0])
    if not rep:
        rep= cmp(p1[4],p2[4])
    return -rep

def cmpRead(r1,r2):
    return cmp(r1[1],r2[1])


cdef int32_t deleteBranch(AsmbGraph graph,list path, int32_t maxlength, int32_t deleted=0):
    
    cdef int32_t   node
    cdef list      parents
    cdef list      sons
    cdef int32_t   i
    
    if len(path) > maxlength:
        return deleted
        

    node = path[-1]
    parents = [y for y in graph.parentIterator(node)]
    if parents :
        for p in parents:                
            sons    = [y for y in graph.neighbourIterator(p)]

            path+=[p]

            if len(sons) < 2:
                deleted = deleteBranch(graph,path,maxlength,deleted)
            else:
                for i in range(len(path)-1,0,-1):
                    try:
                        graph.deleteEdge(path[i],path[i-1])
                        deleted+=1
                    except KeyError:
                        pass
    else:
        for i in range(len(path)-1,0,-1):
            try:
                graph.deleteEdge(path[i],path[i-1])
                deleted+=1
            except KeyError:
                pass
        
    return deleted


cdef set sons(AsmbGraph graph,int node):                          # @DuplicatedSignature
    '''
    Build a set composed of the following reads  
    '''
    return set(graph.neighbourIterator(node))
    
cdef set parents(AsmbGraph graph,int node):                       # @DuplicatedSignature
    '''
    Build a set composed of the previous reads  
    '''
    return set(graph.parentIterator(node))

cdef bint is_junction(AsmbGraph graph,int node):                   # @DuplicatedSignature
    '''
    Predicate testing if a node is a fork or the beginning of the assembly
    '''
    return  (    len(sons(graph,node)) > 1 
              or len(parents(graph,node)) != 1 
            )

cdef bint is_begining(AsmbGraph graph,int node):                   # @DuplicatedSignature
    '''
    Predicate testing if a node is a fork or the beginning of the assembly
    '''
    return  len(parents(graph,node)) == 0 
            

cdef tuple normalizePath(list path, bint *direction):
    cdef int  c = path[0]
    cdef int  r = -path[-1]
    cdef bint d = c < r
    cdef tuple cpath
    
    
    # TODO: verify the property : if more than two nodes are reverse complement then the stem is palindromic 
    if c==r:
        c = path[1]
        r = -path[-2]
        d = c < r
    
    if d:
        cpath=tuple(path)
    else:
        cpath=tuple(-i for i in reversed(path))

    direction[0]=d
    return cpath

cdef bint isPalindrome(list path):
    cdef int lp = len(path)
    cdef int i
    # paths with a odd length cannot be palindromic 
    if (lp & 1) == 1:
        return False
    
    for i in range(lp//2):
        if path[i]!=-path[-i-1]:
            return False
        
    return True

cdef int weight(Assembler assembler,
                list      path,
                bint      palindrome):
    cdef float weight=0
    cdef int x 
    cdef AsmbGraph graph = assembler._graph
    cdef dict nattr
    cdef int nodemax = len(assembler._index) +1 
    
    for x in path:
        if iabs(x) < nodemax:
            try:
                nattr = graph.getNodeAttr(x)
                weight+=nattr.get('coverage',0.0)
            except KeyError:
                pass
           
    weight/= len(path)
    
    if palindrome:
        weight/=2
    
    return <int> (weight * assembler._index.getReadSize())
    
cdef str label(dict stem):
        cdef str label
        cdef int length = stem['length']
        
        if length > 10:
            label="%d : %s->(%d)->%s  [%d]" % (stem['stemid'],
                                               stem['sequence'][0:5].decode('ascii'),
                                               length,
                                               stem['sequence'][-5:].decode('ascii'),
                                               int(stem['weight'])
                                         )
        else:
            label="%d : %s->(%d)  [%d]" % (stem['stemid'],
                                           stem['sequence'].decode('ascii'),
                                           length,
                                           int(stem['weight'])
                                          )
            
        return label


cpdef dict buildstem(Assembler assembler,
               int first,
               int last,
               bytes sequence,
               list path,
               bint circle):
    
    cdef bint palindrome = isPalindrome(path)    
    cdef int w = weight(assembler,path,palindrome)
    cdef int  length = len(sequence)
    cdef dict s = { 'first'      : int(first),
                    'last'       : int(last),
                    'sequence'   : sequence,
                    'length'     : int(length),
                    'path'       : path,
                    'palindrome' : bool(palindrome),
                    'weight'     : int(w),
                    'circle'     : bool(circle),
                    'stemid'     : 0,
                    'label'      : None,
                    'head'       : assembler._index.getRead(first,
                                                            0,
                                                            assembler._index.getReadSize())
                  }
                
    if circle:
        getConfiguration()['orgasm']['logger'].info(" Circle : %6d bp coverage : %6dx" % 
              (length,int(w))) 
        
    return s   

cdef class CompactAssembling(DiGraphMultiEdge):


    def __init__(self,
                 Assembler assembler,
                 bint verbose=True
                ):
        
        cdef int n=0
        cdef int x
        cdef int i                                   # @DuplicatedSignature
        cdef lcontig=0
        cdef int first 
        cdef int last                                # @DuplicatedSignature
        cdef int lseq                                # @DuplicatedSignature
        cdef int lgraph
        cdef int lpath
        cdef bytes sequence                          # @DuplicatedSignature
        cdef dict attr                               # @DuplicatedSignature
        cdef double weight
        cdef double minweight
        cdef AsmbGraph graph
        cdef dict stem
        cdef dict config=getConfiguration()
        
        logger = config['orgasm']['logger']
         
        DiGraphMultiEdge.__init__(self,'compact')
        
        self._paths={}
        self._stemid={}
        self._stemidOk=False
        self._assembler=assembler
        
        graph = assembler._graph
        lgraph=len(graph)


        if verbose:
            logger.info("Compacting graph :")
        else:
            progress = ProgressBar(lgraph,
                                   head="Compacting graph",
                                   seconde=0.1)

        minweight=1000000.
        
        for stem in StemIterator(assembler):
            self.addStem(stem) 
            lcontig+=stem['length']
            weight = stem['weight']
            stem['graphics']={'width':(weight//assembler._index.getReadSize())+1,
                              'arrow':'last'
                             }
            stem['class']='sequence'

            if weight < minweight:
                minweight=weight

            if verbose:
                logger.info(" Stem  : %6d bp (total : %6d) coverage : %6.2f" % (stem['length'],
                                                                                lcontig,
                                                                                weight))
            else:
                progress(lcontig)
           
        self.setStemid()            
             
        logger.info("Minimum stem coverage = %d" % int(minweight))
        
        

    cpdef dict addStem(self, dict stem):
        cdef char[50] buffer1
        cdef char[50] buffer2
        cdef bytes key1 
        cdef bytes key2 
        cdef tuple edge
        cdef list eattr
        cdef int eid
        cdef int first = stem['first']
        cdef int last  = stem['last']
        cdef bint d

        cpath = normalizePath(stem['path'],&d)
        self._paths[cpath] = self._paths.get(cpath,0) + 1
        self._stemidOk=False
        
        snprintf(buffer1,50,b"%d",first)
        key1 = buffer1
        
        snprintf(buffer2,50,b"%d",last)
        key2 = buffer2
        
        edge = (key1,key2)
        
        if edge not in self._edges_attrs:
            self.addNode(first)
            self.addNode(last)
        
            self._nodes[key1][0].add(key2)
            self._nodes[key2][1].add(key1)

            self._edges_attrs[edge]=[]
            
        eattr = self._edges_attrs[edge]
        
        eattr.append(stem)
    
        self._edgecount+=1
                
        return stem

    cpdef deleteEdge(self, int node1, int node2, int edge=-1):
        cdef dict stem              # @DuplicatedSignature
        cdef bint d                 # @DuplicatedSignature
        stem = self.getEdgeAttr(node1,node2,edge)
        cpath = normalizePath(stem['path'],&d)
        self._paths[cpath]-=1
        if self._paths[cpath]==0:
            del self._paths[cpath]
            self._stemidOk=False
        
        DiGraphMultiEdge.deleteEdge(self,node1,node2,edge)
        
    

    cdef int getStemid(self, dict stem) except 0:
        cdef list path = stem['path']
        cdef bint d
        cdef tuple cpath
        cdef int stemid
        cdef list sstems
                        
        if not self._stemidOk:
            sstems=list(self._paths)
            sstems.sort()
            for i in range(len(sstems)):
                self._stemid[sstems[i]]=i+1
            self._stemidOk=True
                
        cpath = normalizePath(path,&d)                
        stemid = self._stemid[cpath]

        if not d:
            stemid=-stemid
                
        return stemid


    def stemIterator(self):
        cdef list stemps
        cdef dict stem
        for stems in self._edges_attrs.values():
            for stem in stems:
                yield stem
        
    cdef void setStemid(self):
        cdef dict stem
    
        for stem in self.stemIterator():
            stem['stemid'] = self.getStemid(stem)
            stem['label']  = label(stem)

    property assembler:
        "A doc string can go here."

        def __get__(self):
            return self._assembler

        
        
    
cdef class StemIterator:

    def __init__(self,Assembler assembler, 
                      bint alllink=False):
        self._assembler=assembler
        self._graph=assembler._graph
        self.edgeName={}
                            
    def __iter__(self):
        cdef NodeIterator ni
        cdef int first
        cdef int n
        cdef set branches
        cdef set brothers
        cdef set fathers
        cdef list lsequence
        cdef int last
        cdef int lseq
        cdef bytes sequence
        cdef list path
        cdef dict stem 
        cdef int lcontig
        cdef AsmbGraph graph =  self._graph
        cdef set junctions = set(int(y) for y in graph.nodeIterator(lambda x: is_junction(graph,x)))
                

        for first in junctions:
            branches=sons(graph,first)
            for n in branches:
                path = [first,n]                               
                brothers=sons(graph,n)
                lsequence=[graph.getEdgeAttr(first,n)['ext']]
                while len(brothers)==1 and n not in junctions:
                    son=brothers.pop()
                    lsequence.append(graph.getEdgeAttr(n,son)['ext'])
                    path.append(son)
                    n=son
                    brothers=sons(graph,n)
    
                last = n
                sequence=b''.join(lsequence)

                stem = buildstem(self._assembler,
                            first,last,
                            sequence,path,
                            False)
                
                lcontig+=stem['length']
                                
                yield stem
                                
                if stem['palindrome']:
                    yield stem
                    
        if lcontig + 2 * len(self.edgeName) < len(graph):
            ccs = graph.connectedComponentIterator()
            
            for cc in ccs:
                circle = not any(is_junction(graph,x) for x in cc)

                if circle:
                    minabs = min(abs(i) for i in cc)
                    print(minabs,file=sys.stderr)
                    if minabs in cc:
                        first = minabs
                        print("in",file=sys.stderr)
                    else:
#                        first = sons(graph,-minabs).pop()
                        first = -minabs
                        print("out",file=sys.stderr)
                    nn = first
                    path=[first]
                    lsequence=[]
                    son=sons(graph,nn).pop()
                    while first != son:
                        path.append(son)
                        lsequence.append(graph.getEdgeAttr(nn,son)['ext'])
                        nn=son
                        son=sons(graph,nn).pop()
    
                    last = first
                    path.append(first)
                    lsequence.append(graph.getEdgeAttr(nn,first)['ext'])
                    
                    sequence=b''.join(lsequence)
                    
                    stem = buildstem(self._assembler,
                                first,last,
                                sequence,
                                path,
                                True)
                
                    yield stem
                                                        
            


cdef class Assembler:
    '''
    totolitoto
    '''

    def __init__(self,Index index, int overlap=90):
        '''
        Create a new assembler object.
        
        :param index: the file name of an read index previously formated using orgasmi binary
        :type index: bytes
        :param seed: a DNA sequenced used as started for the assembly
        :type seed: bytes
        :param overlap: the minimum overlap between two reads
        :type overlap: int
        '''
        cdef bytes starter
        cdef int depth 
        cdef tuple reads
        cdef dict attr
                
        self._graph = AsmbGraph("assembling",index)
        
        self._index = index
        self._overlap=overlap
        self._seeds = []
        self._annotations = {}
                        
        
        
    cpdef tuple readType(self,ids):
        '''
        Internal function : Given a set of read ids, return the one that have to be used
        as standard id. 
         
        The set of ids given to this function corresponds to a set of all strictly identical reads.
          
        :param ids: an iterable elements contining read ids.
        :type ids: iterable
         
        :returns: a tuble of three elements
         
                    - the standard id to use
                    - the length of ids
                    - a set containing all unique ids given as parameter
                     
        :rtype: tuple
        '''
         
        cdef int t 
        cdef int y 
         
        if not ids:
            return None
         
        t = len(self._index)+1
         
        for y in ids:
            if abs(y) < abs(t):
                t = y
                 
        return (t,len(ids),set(ids))
    
    def isEnd(self, int node):
        cdef list s
        cdef list f
        s = [x for x in self.graph.neighbourIterator(node)]
        f = [x for x in self.graph.parentIterator(node)]
        if len(s)==0 and len(f)==1:
            return f[0]
        return 0

    def forkIterator(self):
        
        def has_several_sons(int node):
            return  len(set(self._graph.neighbourIterator(node))) > 1
        
        return self._graph.nodeIterator(has_several_sons)
    
        

           
    def endNodeSet(self, set excluded=None, bint alllink=False):
        if alllink:
            def isfinalnode(int node):
                return len(set(self._graph.neighbourIterator(node)))==0
        else:
            def isfinalnode(int node):                  # @DuplicatedSignature
                return len(set(self._graph.neighbourIterator(node)))==0
        if excluded is None:
            excluded=set()
        return set(x for x in self._graph.nodeIterator(isfinalnode) if x not in excluded)

    def startNodeSet(self, set excluded=None, bint alllink=False):
        if alllink:
            def isstartnode(int node):
                return len(set(self._graph.parentIterator(node)))==0
        else:
            def isstartnode(int node):                  # @DuplicatedSignature
                return len(set(self._graph.parentIterator(node)))==0
            
        if excluded is None:
            excluded=set()
        return set(x for x in self._graph.nodeIterator(isstartnode) if x not in excluded)

    def cleanDeadBranches(self,int32_t maxlength=10, bint alllink=False):

        cdef AsmbGraph graph
        cdef int32_t   d
        cdef int32_t   cd
        cdef int32_t   i                                # @DuplicatedSignature
        cdef set       ep
        cdef set       sp
                   
        graph=self.graph

        cd=0    
        d=1
            
        if sys.stderr.isatty():
            print('',file=sys.stderr)
        
        while d > 0:
            d=0
            
            for i in self.endNodeSet(alllink=alllink):
                if sys.stderr.isatty():
                    print("Remaining edges : %d node : %d" % (graph.edgeCount(),len(graph)),
                          end='\r',
                          file=sys.stderr)
                d+=deleteBranch(graph,[i],maxlength)
            cd+=d
            
        
        ep = self.endNodeSet(alllink=alllink)
        sp = self.startNodeSet(alllink=alllink)
        
        for i in ep & sp:
            try:
                graph.deleteNode(i)
            except KeyError:
                pass
            
        if sys.stderr.isatty():
            print ("Remaining edges : %d node : %d" % (graph.edgeCount(),len(graph)),
                   end='\r',
                   file=sys.stderr)

        return cd            
                
    def __len__(self):
        return self._graph.nodeCount()
            
    def compactAssembling(self, bint verbose=True):
        return CompactAssembling(self,verbose)
    
    
    property graph:

        "A doc string can go here."

        def __get__(self):
            return self._graph


    property seeds:

        "A doc string can go here."

        def __get__(self):                              # @DuplicatedSignature
            return self._seeds


    property index:

        "A doc string can go here."

        def __get__(self):                              # @DuplicatedSignature
            return self._index



                       



#cython: language_level=3

'''
Created on 27 mars 2016

@author: coissac
'''

import sys
from ..utils import bytes2str,str2bytes
from .config    cimport getConfiguration 


cdef class ProgressBar:
    cdef clock_t clock(self):
        cdef clock_t t
        cdef timeval tp
        cdef clock_t s
        
        <void> gettimeofday(&tp,NULL)
        s = <clock_t> (<double> tp.tv_usec * 1.e-6 * <double> CLOCKS_PER_SEC)
        t = tp.tv_sec * CLOCKS_PER_SEC + s 
        
        return t

    def __init__(self,
                 off_t maxi,
                 dict  config={},
                 str head="",
                 double seconde=0.1):
        self.starttime = self.clock()
        self.lasttime  = self.starttime
        self.tickcount = <clock_t> (seconde * CLOCKS_PER_SEC)
        self.freq      = 1
        self.cycle     = 0
        self.arrow     = 0
        self.lastlog   = 0
        
        
        self.ontty = sys.stderr.isatty()
        
        if (maxi<=0):
            maxi=1
            
        self.maxi = maxi
        self._head = str2bytes(head)
        self.chead= self._head 
        
        
        try:
            if not config:
                config=getConfiguration()
                
            self.logger=config[config["__root_config__"]]["logger"]
        except RuntimeError:
            self.logger=None

        self.wheel =  '|/-\\'
        self.spaces='          ' \
                    '          ' \
                    '          ' \
                    '          ' \
                    '          '
        self.diese ='##########' \
                    '##########' \
                    '##########' \
                    '##########' \
                    '##########'  
                           
    def __call__(self,object pos):
        cdef off_t    ipos
        cdef clock_t  elapsed
        cdef clock_t  newtime
        cdef clock_t  delta
        cdef clock_t  more 
        cdef double   percent 
        cdef tm remain
        cdef int days,hour,minu,sec
        cdef off_t fraction
        cdef int tenth
        
        self.cycle+=1
    
        if self.cycle % self.freq == 0:
            self.cycle=1
            newtime  = self.clock()
            delta         = newtime - self.lasttime
            self.lasttime = newtime
            elapsed       = newtime - self.starttime
#            print(" ",delta,elapsed,elapsed/CLOCKS_PER_SEC,self.tickcount)
            
            if   delta < self.tickcount / 5 :
                self.freq*=2
            elif delta > self.tickcount * 5 and self.freq>1:
                self.freq/=2
                
            
            if callable(pos):
                ipos=pos()
            else:
                ipos=pos
                
            if ipos==0:
                ipos=1                

            percent = <double>ipos/<double>self.maxi
            more = <time_t>((<double>elapsed / percent * (1. - percent))/CLOCKS_PER_SEC)
            <void>gmtime_r(&more, &remain)
            days  = remain.tm_yday 
            hour  = remain.tm_hour
            minu  = remain.tm_min
            sec   = remain.tm_sec
                
            if self.ontty:
                fraction=<int>(percent * 50.)
                self.arrow=(self.arrow+1) % 4
            
                if days:
                    <void>fprintf(stderr,b'\r%s %5.1f %% |%.*s%c%.*s] remain : %d days %02d:%02d:%02d',
                                    self.chead,
                                    percent*100,
                                    fraction,self.diese,
                                    self.wheel[self.arrow],
                                    50-fraction,self.spaces,
                                    days,hour,minu,sec)
                else:
                    <void>fprintf(stderr,b'\r%s %5.1f %% |%.*s%c%.*s] remain : %02d:%02d:%02d',
                                    self.chead,
                                    percent*100.,
                                    fraction,self.diese,
                                    self.wheel[self.arrow],
                                    50-fraction,self.spaces,
                                    hour,minu,sec)

            tenth = int(percent * 10)
            if tenth != self.lastlog and not self.ontty and self.logger is not None:
                
#                if self.ontty:
#                    <void>fputs(b'\n',stderr)
                                   
                self.logger.info('%s %5.1f %% remain : %02d:%02d:%02d' % (
                                        bytes2str(self._head),
                                        percent*100.,
                                        hour,minu,sec))
                self.lastlog=tenth
        else:
            self.cycle+=1

    property head:
    
        def __get__(self):
            return self._head
        
        def __set__(self,str value):
            self._head=str2bytes(value)
            self.chead=self._head



'''
Created on 22 janv. 2016

@author: coissac
'''

import sys
from urllib import request
import os.path

from obidistutils.serenity.util import get_serenity_dir
from obidistutils.serenity.rerun import rerun_with_anothe_python
from obidistutils.serenity.checkpython import is_a_virtualenv_python

getpipurl="https://bootstrap.pypa.io/get-pip.py"

def bootstrap():

    getpipfile=os.path.join(get_serenity_dir(),"get-pip.py")
    
    with request.urlopen(getpipurl) as getpip:
        with open(getpipfile,"wb") as out:
            for l in getpip:
                out.write(l)
                
    python = sys.executable
    
    if is_a_virtualenv_python():
        command= "%s %s" % (python,getpipfile)        
    else:
        command= "%s %s --user" % (python,getpipfile)
    
    os.system(command)
    
    rerun_with_anothe_python(python)
    


'''
Created on 20 oct. 2012

@author: coissac
'''

# try:
#     from setuptools.command.install_scripts import install_scripts as ori_install_scripts
# except ImportError:
#     from distutils.command.install_scripts import install_scripts as ori_install_scripts

from distutils.command.install_scripts import install_scripts as ori_install_scripts

import os.path
from distutils import log

class install_scripts(ori_install_scripts):

    def initialize_options(self):
        ori_install_scripts.initialize_options(self)
        self.public_dir = None
                        
                    
    def install_public_link(self):
        self.mkpath(self.public_dir)
        for file in self.get_outputs():
            log.info("exporting file %s -> %s", file,os.path.join(self.public_dir,
                                    os.path.split(file)[1]
                                    ))
            if not self.dry_run:
                dest = os.path.join(self.public_dir,
                                    os.path.split(file)[1]
                                   )
                if os.path.exists(dest):
                    os.unlink(dest)
                os.symlink(file,dest)
                    


    def run(self):
        ori_install_scripts.run(self)
        if self.distribution.serenity:
            self.public_dir=os.path.join(self.install_dir,"../export/bin")
            self.public_dir=os.path.abspath(self.public_dir)
            self.install_public_link()





# cython: language_level=3
from asyncio.log import logger
cimport cython

from ._orgasm cimport *
from cpython.bytes cimport PyBytes_Size,PyBytes_AsString,PyBytes_GET_SIZE

from cpython.dict cimport PyDict_Merge,PyDict_GetItem
from cpython.list cimport PyList_GET_ITEM,PyList_SetItem,PyList_Append
from cpython.int  cimport PyInt_AS_LONG,PyInt_FromLong

from array import array

from orgasm.backtranslate._ahocorasick cimport *
from orgasm.utils.dna cimport isDNA
from orgasm.apps.progress cimport ProgressBar
#from time import time

from posix.time cimport timeval,gettimeofday

import sys

cdef bint adapter(bytes probe, tuple adapterSeq5, tuple adapterSeq3):
    return ( any([probe.startswith(y) for y in adapterSeq5]) or 
             any([probe.endswith(y) for y in adapterSeq3]) )


cdef extern from "stdlib.h":
    ctypedef unsigned int size_t
    int32_t abs(int32_t i)

cdef extern from "math.h":
    double log10(double x)
    double ceil(double x)

cdef extern from "orgasm.h":
    void freeBuffer(buffer_t *buffer)

    buffer_t *loadIndexedReads(char *indexname)

    int32_t lookForString(buffer_t *buffer, char *key, size_t length, int32_t* endoflist)
    int32_t fastLookForString(buffer_t *buffer, char *key, size_t length, int32_t* fcount, int32_t* rcount, int32_t* fpos, int32_t* rpos)
    int32_t nextRead(buffer_t *buffer, int32_t current, size_t length, int32_t* endoflist)
    int32_t lookForReadsIds(buffer_t *buffer, int32_t key, int32_t* endoflist)
    int32_t nextReadIds(buffer_t *buffer, int32_t current, int32_t* endoflist)
    char *getRead(buffer_t *buffer, int32_t recordid, uint32_t begin, int32_t length, char *dest)

cdef bint lowcomplexity(bytes s):
    """
    Returns True if the word s is an homopolymer, an homo-dimer
    or an homo-trimer

    @param s: the string to test
    @type s: bytes

    @return True if s is an homopolymer, an homedimer
            or an homotrimer, False otherwise
    @rtype: bool
    """
    cdef int ls=PyBytes_GET_SIZE(s)
    cdef char* ps = PyBytes_AsString(s)
    cdef bint r = (strncmp(ps,ps+2,ls - 2) * strncmp(ps,ps+3,ls - 3)) == 0
    return r 
#    return s[0:-2] == s[2:] or s[0:-3] == s[3:]

cdef double time():
    cdef double  seconds
    cdef double  usec
    cdef timeval tv
    gettimeofday (&tv, NULL);
    seconds = tv.tv_sec
    usec    = tv.tv_usec / 1000000.
    return seconds + usec 
    
cdef class Index:

    def __init__(self,str name):
        cdef int minfake

        self._index=loadIndexedReads(bytes(name,'latin1'))
        minfake = <int>(10**ceil(log10(self._index.readCount))) + 1
        self._fakes=FakeReads(minfake,self._index.readSize)

    cpdef bint contains(self, bytes word):
        cdef int32_t endoflist
        cdef size_t  lword = PyBytes_Size(word)

        lookForString(self._index,word,lword,&endoflist)

        return endoflist==0

    cpdef int count(self, bytes word):
        return len(self.lookForString(word))

    cpdef int len(self):
        return self._index.readCount

    cpdef bytes getRead(self, int32_t readid, int32_t begin, int32_t length):
        """
        getRead methods allows to access to the the sequence of the reads stored in the
        index. The read id is include in the interval [-n,-1] U [1,n] with n the count
        of read in the index. the read 0 does not exist.
        """
        cdef char buffer[500]
        cdef bytes seq

        assert begin < self._index.readSize,"Begin must be smaller than read size"
        assert begin+length <= self._index.readSize,"Begin+Length must be smaller than read size"

        if self._fakes.isFake(readid):
            seq = self._fakes.getRead(readid,begin,length)
        elif readid==0 or abs(readid)>self._index.readCount:
            raise IndexError(readid)
        else:
            seq=getRead(self._index, readid, begin, length, buffer)

        return seq

    cpdef dict getExtensions(self, bytes word):
        cdef dict rep
        cdef int32_t readid                                 # @DuplicatedSignature
        cdef size_t  lword = PyBytes_Size(word)                      # @DuplicatedSignature
        cdef size_t  lext  = self._index.readSize - lword
        cdef bytes e
        cdef list  data
        cdef c_array.array w
        cdef long[:] cw
        cdef size_t  lw
        cdef size_t  i
        cdef double ellapsed = time()
        cdef PyObject *pdata
        cdef object o
        cdef list ids
        cdef int nread

        rep = {}
        
        w = self.lookForString(word)
        lw = len(w)
        cw = w
        
        for i in range(lw):
            readid = cw[i]
            # assert abs(readid) < self._index.readCount
            e = self.getRead(readid,lword, lext)
            pdata = PyDict_GetItem(rep,e)
            if pdata==NULL:
                data=[0,[]]
                rep[e]=data
            else:
                data= <list> pdata

            data[0]+=1

            pdata = PyList_GET_ITEM(data,1)
            ids = <list> pdata
            PyList_Append(ids,PyInt_FromLong(readid))
            
        ellapsed = time()  - ellapsed
        
        if ellapsed > 0.5:
            print("\nWarning too long getExtension (%4.2fs): %s -> %d matches\n" % (ellapsed,word,len(w)),file=sys.stderr)

        return rep

    cpdef dict checkedExtensions(self, bytes probe,
                                       tuple adapters5=(),
                                       tuple adapters3=(),
                                       int minread=20,
                                       double minratio=0.1,
                                       int mincov=1,
                                       int minoverlap=40,
                                       int extlength=1,
                                       bint lowfilter=True,
                                       object restrict=None,
                                       bint exact=True):
        '''
        :param probe: The sequence used as prefix in the selection
        :type probe:  bytes
        :param minread: the minimum count of read to consider
        :type minread:  int
        :param minratio: minimum ratio between occurrences of an extension
                         and the occurrences of the most frequent extension
                         to keep it.
        :type minratio:  float [0,1]
        :param mincov:   minimum occurrences of an extension to keep it
        :type mincov:    int
        :param minoverlap: minimum length of the probe usable
        :type minoverlap: int
        '''

        cdef dict  ext       = {}
        cdef dict  lext
        cdef int   covmax    = 0
        cdef int   readcount = 0
        cdef list  e,rs,ks,rk
        cdef list  allexts
        cdef dict  extcount = {}
        cdef bytes k
        cdef int   x
        cdef int   r,rn,rsize
        cdef bytes oriprobe

        oriprobe = probe
        
            
        
        while ( not (     (    lowcomplexity(probe) 
                            or adapter(probe, adapters5,adapters3)
                          ) 
                      and      lowfilter
                    ) 
                and PyBytes_Size(probe) >= minoverlap   
                and (    readcount < minread 
                      or exact
                    )
              ):
            
            trueoverlap = len(probe)
            lext = self.getExtensions(probe)
                        
            if restrict is not None:
                ks = list(lext.keys())
                for k in ks:
                    rs=lext[k][1]
                    rk=[]
                    for r in rs:
                        if r in restrict:
                            rk.append(r)
                        else:
                            rn = self.getPairedRead(r)
                            if rn:
                                rn = self.getIds(rn)[0]
                                if rn in restrict:
                                    rk.append(r)
                            
                    if rk:
                        lext[k][1]=rk
                    else:
                        del lext[k]
                                    
#             if restrict is not None:
#                 ks = lext.keys()
#                 for k in ks:
#                     rs=lext[k][1]
#                     rk=[]
#                     for r in rs:
#                         rn = self.getPairedRead(r)
#                         if rn:
#                             rn = self.getIds(rn)[0]
#                             if rn in restrict:
#                                 rk.append(r)
#                     if rk:
#                         lext[k][1]=rk
#                     else:
#                         del lext[k]

            if lext:
                for e in lext.itervalues():
                    readcount+=e[0]
                ext.update(lext)
                
            probe=probe[1:]
            
            
        ##-------------------END OF THE WHILE LOOP ----------------------##

        if readcount < minread:
            ext={}
        elif ext:
            allexts = [(k[0:extlength],ext[k][0]) for k in ext.keys()]
            extcount= {}
            for k,x in allexts:
                extcount[k]=extcount.get(k,0)+x
                
            rsize = self.getReadSize()
            
            # Evaluate the true coverage according the actual overlap
            for k,x in extcount.items():
                extcount[k]=x * rsize // (rsize - trueoverlap)

            covmax = max(extcount.values())

            it = list(extcount.items())

            for k,x in it:
                if (   x < mincov 
                    or <double>x/<double>covmax < minratio):
                    del extcount[k]
                else:
                    data = ext.get(k,(0,[]))[1]
                    if not data:
                        data=list(self.getReadIds(oriprobe+k)[2])

                    extcount[k]=(extcount[k],data)


        return extcount


    cpdef int getReadSize(self):
        return self._index.readSize

    cpdef tuple getReadIds(self, bytes seq):

        cdef c_array.array x
        cdef int  y
        cdef int  t

        assert len(seq)==self._index.readSize,"seq must have the same length than reads"

        x = self.lookForString(seq)

        if x:
            t = len(self)+1

            for y in x:
                if abs(y) < abs(t):
                    t = y
            y = len(x)
        else:
            f = self._fakes.getReadIds(seq)
            return f 

        return (t,y,set(x))

    cpdef int32_t getPairedRead(self, int32_t id):
        cdef int32_t iid = abs(id)
        cdef int32_t pid

        if iid > self._index.readCount:
            pid = 0
        else:
            pid = self._index.order2[iid-1] + 1

        if id > 0:
            pid = - pid

        return pid

    cpdef tuple normalizedPairedEndsReads(self, int32_t id):
        '''
        Returns for a read id the normalized set of paired reads corresponting to it.

        :param id: the read id
        @return: a tuple of two elements
                    - an integer value : the normalized id
                    - a list of integers : the normalized ids of the paired reads
        '''
        cdef int32_t nid
        cdef int32_t n
        cdef set synomymes
        cdef int32_t i
        cdef list paired
        cdef tuple k
        cdef int idsign = 1 if id > 0 else -1 
        
        nid,n,synomymes = self.getIds(id)
          
        if id not in synomymes:
            nid=-nid
            synomymes=set([-i for i in synomymes])

        goodsyno = [x for x in synomymes if x * idsign > 0]

        paired = [self.getPairedRead(i) for i in goodsyno]
        
        paired = [(j,self.getIds(j)) for j in paired]
        paired = [k[0] if j in k[2] else -k[0] for j,k in paired]
        return nid,paired


    cpdef dict lookForSeeds(self, dict sequences, 
                            int kup=-1, int mincov=1,
                            float identity=0.5,
                            object logger=None):

        cdef AhoCorasick patterns
        cdef dict matches
        cdef str k   
        cdef bint  nuc   
        cdef ProgressBar progress
        cdef int i
        
        nuc = all([isDNA(sequences[k]) for k in sequences])
            
        if nuc:
            if logger is not None:
                logger.info('Matching against nucleic probes')
            patterns = NucAhoCorasick()
            kup = 12 if kup < 0 else kup
        else:
            if logger is not None:
                logger.info('Matching against protein probes')
            patterns = ProtAhoCorasick()
            kup = 4 if kup < 0 else kup

        progress = ProgressBar(len(sequences),
                               head="Building Aho-Corasick automata",
                               seconde=0.1)
        i=0
        for k in sequences:
            patterns.addSequence(sequences[k],k,kup)
            i=i+1
            progress(i)

        
        patterns.finalize()
                
        #minmatch = 50 if nuc else 15
        minmatch = int((self.getReadSize() // (1 if nuc else 3)) * identity)
        logger.info('Minimum word matches = %d ' % minmatch)
        
        matches = patterns.scanIndex(self,minmatch,-1,mincov)

        return matches

    cpdef bint isFake(self,int32_t id):
        return self._fakes.isFake(id)

    cpdef tuple getIds(self, int32_t id):
        cdef int32_t endoflist                                      # @DuplicatedSignature
        cdef int32_t readid                                         # @DuplicatedSignature
        cdef list    reads = []
        cdef int  y                                                 # @DuplicatedSignature
        cdef int  x
        cdef int  t                                                 # @DuplicatedSignature

        if id==0:
            raise IndexError(id)

        if self._fakes.isFake(id):
            return self._fakes.getIds(id)

        y = abs(id)
        
        if y > self._index.readCount:
            raise IndexError("%d (Maximum id is %d)" % (id,self._index.readCount))

        readid = lookForReadsIds(self._index,id,&endoflist)

        while(endoflist==0):
            reads.append(readid)
            readid = nextReadIds(self._index,readid,&endoflist)

        if reads:
            t = len(self)+1

            for y in reads:
                if abs(y) < abs(t):
                    t = y

            if t < 0:
                reads = [-x for x in reads]
                t = -t

        return (t,len(reads),set(reads))


    def __dealloc__(self):
        freeBuffer(self._index)

    def __contains__(self, bytes word):
        return self.contains(word)
    
    @cython.boundscheck(False)
    cdef c_array.array lookForString(self, bytes word):
        cdef int32_t fpos                                       # @DuplicatedSignature
        cdef int32_t rpos                                       # @DuplicatedSignature
        cdef int32_t fcount                                     # @DuplicatedSignature
        cdef int32_t rcount                                     # @DuplicatedSignature
        cdef int32_t count                                      # @DuplicatedSignature
        cdef int32_t readid                                     # @DuplicatedSignature
        cdef size_t  lword = len(word)                          # @DuplicatedSignature
        cdef c_array.array readids = array('l', [0])
        cdef long[:] creadids

        count  = fastLookForString(self._index,word,lword,&fcount,&rcount,&fpos,&rpos)
        c_array.resize(readids,count)
        creadids=readids
#        print fcount,rcount,fpos,rpos
#        print count
        for readid in range(count):
            if readid < fcount:
                fpos+=1
                creadids[readid]=fpos
            else:
                creadids[readid]=-(<int32_t>self._index.order1[rpos])-1
                rpos+=1
                
#         readids=list(range(fpos+1,fpos+count+1))
#         for readid in range(fcount,count):
#             readids[readid]=-(<int32_t>self._index.order1[readid-fcount+rpos])-1 
            
#        print readids
        
        return readids

    def __getitem__(self, bytes word):
        return self.lookForString(word)
    

    def __len__(self):
        return self.len()

    def checkReverseOrder(self):
        conv = {'A':'0','C':'1','T':'2','G':'3'}
        from orgasm.utils.dna import reverseComplement
        olds=" "
        for i in range(len(self)):
            j = self._index.order1[i] + 1
            s = "".join(conv[x] for x in reverseComplement(self.getRead(j,0,self.getReadSize())))
            if olds > s:
                print(i,olds,s)
                raise AssertionError
            else:
                olds=s



    property fakes:

        "An direct access to the fake reads."

        def __get__(self):
            return self._fakes












#cython: language_level=3

'''
Created on 30 mars 2016

@author: coissac
'''

cdef class LineBuffer:

    def __init__(self,object fileobj,int size=100000000):
        self.fileobj=fileobj 
        self.size=size
    
    def __iter__(self):
        cdef list buff = self.fileobj.readlines(self.size)
        cdef str  l
        
        while buff:
            for l in buff:
                yield l
            buff = self.fileobj.readlines(self.size)
    
